from typing import Type

from maimai_ffi.exceptions import AimeServerError, ArcadeIdentifierError, TitleServerBlockedError, TitleServerNetworkError

AimeServerError: Type[Exception]
ArcadeIdentifierError: Type[Exception]
TitleServerBlockedError: Type[Exception]
TitleServerNetworkError: Type[Exception]


class MaimaiPyError(Exception):
    """Base exception class for all exceptions raised by maimai_py."""


class InvalidJsonError(MaimaiPyError):
    """Invalid JSON response from the provider."""


class InvalidPlayerIdentifierError(MaimaiPyError):
    """Player identifier is invalid for the provider.

    For example, friend code is not applicable for Diving Fish provider, the username is not applicable for LXNS provider.

    Also, if the player is not found on that provider, this exception will be raised.
    """


class InvalidDeveloperTokenError(MaimaiPyError):
    """Developer token is not provided or token is invalid."""


class InvalidPlateError(MaimaiPyError):
    """Provided version or plate is invalid.

    Plate should be formatted as two/three characters (version + kind), e.g. "桃将", "舞舞舞"

    The following versions are valid:

    霸, 舞, 初, 真, 超, 檄, 橙, 晓, 桃, 樱, 紫, 堇, 白, 雪, 辉, 熊, 华, 爽, 煌, 星, 宙, 祭, 祝, 双, 宴.

    The following kinds are valid:

    将, 者, 極, 极, 舞舞, 神

    """


class PrivacyLimitationError(MaimaiPyError):
    """The user has not accepted the privacy policy or exceeded the privacy limit of the provider."""


class InvalidWechatTokenError(MaimaiPyError):
    """Wahlap Wechat OffiAccount token is invalid or expired."""


WechatTokenExpiredError = InvalidWechatTokenError



from .arcade import ArcadeProvider
from .base import (
    IAliasProvider,
    IAreaProvider,
    ICurveProvider,
    IItemListProvider,
    IPlayerIdentifierProvider,
    IPlayerProvider,
    IProvider,
    IRegionProvider,
    IScoreProvider,
    IScoreUpdateProvider,
    ISongProvider,
)
from .divingfish import DivingFishProvider
from .local import LocalProvider
from .lxns import LXNSProvider
from .wechat import WechatProvider
from .yuzu import YuzuProvider

__all__ = [
    "IProvider",
    "IScoreUpdateProvider",
    "IAliasProvider",
    "IAreaProvider",
    "ICurveProvider",
    "IItemListProvider",
    "IPlayerProvider",
    "IRegionProvider",
    "IScoreProvider",
    "ISongProvider",
    "IPlayerIdentifierProvider",
    "ArcadeProvider",
    "DivingFishProvider",
    "LocalProvider",
    "LXNSProvider",
    "WechatProvider",
    "YuzuProvider",
]



import asyncio
import dataclasses
import hashlib
import warnings
from collections import defaultdict
from functools import cached_property
from typing import Any, AsyncGenerator, Generic, Iterable, Literal, Optional, Type, TypeVar

from aiocache import BaseCache, SimpleMemoryCache
from httpx import AsyncClient

from maimai_py.models import *
from maimai_py.providers import *
from maimai_py.utils import UNSET, _UnsetSentinel

PlayerItemType = TypeVar("PlayerItemType", bound=PlayerItem)

T = TypeVar("T", bound=Score)


class MaimaiItems(Generic[PlayerItemType]):
    _client: "MaimaiClient"
    _namespace: str

    def __init__(self, client: "MaimaiClient", namespace: str) -> None:
        """@private"""
        self._client = client
        self._namespace = namespace

    async def _configure(self, provider: Union[IItemListProvider, _UnsetSentinel] = UNSET) -> "MaimaiItems":
        cache_obj, cache_ttl = self._client._cache, self._client._cache_ttl
        # Check if the provider is unset, which means we want to access the cache directly.
        if isinstance(provider, _UnsetSentinel):
            if await cache_obj.get("provider", None, namespace=self._namespace) is not None:
                return self
        # Really assign the unset provider to the default one.
        provider = LXNSProvider() if PlayerItemType in [PlayerIcon, PlayerNamePlate, PlayerFrame] else LocalProvider()
        # Check if the current provider hash is different from the previous one, which means we need to reconfigure.
        current_provider_hash = provider._hash()
        previous_provider_hash = await cache_obj.get("provider", "", namespace=self._namespace)
        # If different or previous is empty, we need to reconfigure the items.
        if current_provider_hash != previous_provider_hash:
            val: dict[int, Any] = await getattr(provider, f"get_{self._namespace}")(self._client)
            await asyncio.gather(
                cache_obj.set("provider", current_provider_hash, ttl=cache_ttl, namespace=self._namespace),  # provider
                cache_obj.set("ids", [key for key in val.keys()], namespace=self._namespace),  # ids
                cache_obj.multi_set(val.items(), namespace=self._namespace),  # items
            )
        return self

    async def get_all(self) -> list[PlayerItemType]:
        """All items as list.

        This method will iterate all items in the cache, and yield each item one by one. Unless you really need to iterate all items, you should use `by_id` or `filter` instead.

        Returns:
            A list with all items in the cache, return an empty list if no item is found.
        """
        item_ids: Optional[list[int]] = await self._client._cache.get("ids", namespace=self._namespace)
        assert item_ids is not None, f"Items not found in cache {self._namespace}, please call configure() first."
        return await self._client._multi_get(item_ids, namespace=self._namespace)

    async def get_batch(self, ids: Iterable[int]) -> list[PlayerItemType]:
        """Get items by their IDs.

        Args:
            ids: the IDs of the items.
        Returns:
            A list of items if they exist, otherwise return an empty list.
        """
        return await self._client._multi_get(ids, namespace=self._namespace)

    async def by_id(self, id: int) -> Optional[PlayerItemType]:
        """Get an item by its ID.

        Args:
            id: the ID of the item.
        Returns:
            the item if it exists, otherwise return None.
        """
        return await self._client._cache.get(id, namespace=self._namespace)

    async def filter(self, **kwargs) -> list[PlayerItemType]:
        """Filter items by their attributes.

        Ensure that the attribute is of the item, and the value is of the same type. All conditions are connected by AND.

        Args:
            kwargs: the attributes to filter the items by.
        Returns:
            an async generator yielding items that match all the conditions, yields no items if no item is found.
        """
        cond = lambda item: all(getattr(item, key) == value for key, value in kwargs.items() if value is not None)
        return [item for item in await self.get_all() if cond(item)]


class MaimaiSongs:
    _client: "MaimaiClient"

    def __init__(self, client: "MaimaiClient") -> None:
        """@private"""
        self._client = client

    async def _configure(
        self,
        provider: Union[ISongProvider, _UnsetSentinel],
        alias_provider: Union[IAliasProvider, None, _UnsetSentinel],
        curve_provider: Union[ICurveProvider, None, _UnsetSentinel],
    ) -> "MaimaiSongs":
        cache_obj, cache_ttl = self._client._cache, self._client._cache_ttl
        # Check if all providers are unset, which means we want to access the cache directly.
        if isinstance(provider, _UnsetSentinel) and isinstance(alias_provider, _UnsetSentinel) and isinstance(curve_provider, _UnsetSentinel):
            if await cache_obj.get("provider", None, namespace="songs") is not None:
                return self
        # Really assign the unset providers to the default ones.
        provider = provider if not isinstance(provider, _UnsetSentinel) else LXNSProvider()
        alias_provider = alias_provider if not isinstance(alias_provider, _UnsetSentinel) else YuzuProvider()
        curve_provider = curve_provider if not isinstance(curve_provider, _UnsetSentinel) else None  # Don't fetch curves if not provided.
        # Check if the current provider hash is different from the previous one, which means we need to reconfigure.
        current_provider_hash = hashlib.md5(
            (provider._hash() + (alias_provider._hash() if alias_provider else "") + (curve_provider._hash() if curve_provider else "")).encode()
        ).hexdigest()
        previous_provider_hash = await cache_obj.get("provider", "", namespace="songs")
        # If different or previous is empty, we need to reconfigure the songs.
        if current_provider_hash != previous_provider_hash:
            # Get the resources from the providers in parallel.
            songs, song_aliases, song_curves = await asyncio.gather(
                provider.get_songs(self._client),
                alias_provider.get_aliases(self._client) if alias_provider else asyncio.sleep(0, result={}),
                curve_provider.get_curves(self._client) if curve_provider else asyncio.sleep(0, result={}),
            )

            # Build the song objects and set their aliases and curves if provided.
            for song in songs:
                if alias_provider is not None and (aliases := song_aliases.get(song.id, None)):
                    song.aliases = aliases
                if curve_provider is not None:
                    if curves := song_curves.get((song.id, SongType.DX), None):
                        diffs = song.get_difficulties(SongType.DX)
                        [diff.__setattr__("curve", curves[i]) for i, diff in enumerate(diffs) if i < len(curves)]
                    if curves := song_curves.get((song.id, SongType.STANDARD), None):
                        diffs = song.get_difficulties(SongType.STANDARD)
                        [diff.__setattr__("curve", curves[i]) for i, diff in enumerate(diffs) if i < len(curves)]
                    if curves := song_curves.get((song.id, SongType.UTAGE), None):
                        diffs = song.get_difficulties(SongType.UTAGE)
                        [diff.__setattr__("curve", curves[i]) for i, diff in enumerate(diffs) if i < len(curves)]

            # Set the cache with the songs, aliases, and versions.
            await asyncio.gather(
                cache_obj.set("provider", current_provider_hash, ttl=cache_ttl, namespace="songs"),  # provider
                cache_obj.set("ids", [song.id for song in songs], namespace="songs"),  # ids
                cache_obj.multi_set(iter((song.id, song) for song in songs), namespace="songs"),  # songs
                cache_obj.multi_set(iter((song.title, song.id) for song in songs), namespace="tracks"),  # titles
                cache_obj.multi_set(iter((li, id) for id, ul in song_aliases.items() for li in ul), namespace="aliases"),  # aliases
                cache_obj.set(
                    "versions",
                    {f"{song.id} {diff.type} {diff.level_index}": diff.version for song in songs for diff in song.get_difficulties()},
                    namespace="songs",
                ),  # versions
            )
        return self

    async def get_all(self) -> list[Song]:
        """All songs as list.

        This method will iterate all songs in the cache, and yield each song one by one. Unless you really need to iterate all songs, you should use `by_id` or `filter` instead.

        Returns:
            A list of all songs in the cache, return an empty list if no song is found.
        """
        song_ids: Optional[list[int]] = await self._client._cache.get("ids", namespace="songs")
        assert song_ids is not None, "Songs not found in cache, please call configure() first."
        return await self._client._multi_get(song_ids, namespace="songs")

    async def get_batch(self, ids: Iterable[int]) -> list[Song]:
        """Get songs by their IDs.

        Args:
            ids: the IDs of the songs.
        Returns:
            A list of songs if they exist, otherwise return an empty list.
        """
        return await self._client._multi_get(ids, namespace="songs")

    async def by_id(self, id: int) -> Optional[Song]:
        """Get a song by its ID.

        Args:
            id: the ID of the song, always smaller than `10000`, should (`% 10000`) if necessary.
        Returns:
            the song if it exists, otherwise return None.
        """
        return await self._client._cache.get(id, namespace="songs")

    async def by_title(self, title: str) -> Optional[Song]:
        """Get a song by its title.

        Args:
            title: the title of the song.
        Returns:
            the song if it exists, otherwise return None.
        """
        song_id = await self._client._cache.get(title, namespace="tracks")
        song_id = 383 if title == "Link(CoF)" else song_id
        return await self._client._cache.get(song_id, namespace="songs") if song_id else None

    async def by_alias(self, alias: str) -> Optional[Song]:
        """Get song by one possible alias.

        Args:
            alias: one possible alias of the song.
        Returns:
            the song if it exists, otherwise return None.
        """
        if song_id := await self._client._cache.get(alias, namespace="aliases"):
            if song := await self._client._cache.get(song_id, namespace="songs"):
                return song

    async def by_artist(self, artist: str) -> list[Song]:
        """Get songs by their artist, case-sensitive.

        Args:
            artist: the artist of the songs.
        Returns:
            an async generator yielding songs that match the artist.
        """
        return [song for song in await self.get_all() if song.artist == artist]

    async def by_genre(self, genre: Genre) -> list[Song]:
        """Get songs by their genre, case-sensitive.

        Args:
            genre: the genre of the songs.
        Returns:
            an async generator yielding songs that match the genre.
        """
        return [song for song in await self.get_all() if song.genre == genre]

    async def by_bpm(self, minimum: int, maximum: int) -> list[Song]:
        """Get songs by their BPM.

        Args:
            minimum: the minimum (inclusive) BPM of the songs.
            maximum: the maximum (inclusive) BPM of the songs.
        Returns:
            an async generator yielding songs that match the BPM.
        """
        return [song for song in await self.get_all() if minimum <= song.bpm <= maximum]

    async def by_versions(self, versions: Version) -> list[Song]:
        """Get songs by their versions, versions are fuzzy matched version of major maimai version.

        Args:
            versions: the versions of the songs.
        Returns:
            an async generator yielding songs that match the versions.
        """
        cond = lambda song: versions.value <= song.version < all_versions[all_versions.index(versions) + 1].value
        return [song for song in await self.get_all() if cond(song)]

    async def by_keywords(self, keywords: str) -> list[Song]:
        """Get songs by their keywords, keywords are matched with song title, artist and aliases.

        Args:
            keywords: the keywords to match the songs.
        Returns:
            an async generator yielding songs that match the keywords.
        """
        cond = lambda song: keywords.lower() in f"{song.title} + {song.artist} + {''.join(a for a in (song.aliases or []))}".lower()
        return [song for song in await self.get_all() if cond(song)]

    async def filter(self, **kwargs) -> list[Song]:
        """Filter songs by their attributes.

        Ensure that the attribute is of the song, and the value is of the same type. All conditions are connected by AND.

        Args:
            kwargs: the attributes to filter the songs by.
        Returns:
            an async generator yielding songs that match all the conditions.
        """
        cond = lambda song: all(getattr(song, key) == value for key, value in kwargs.items() if value is not None)
        return [song for song in await self.get_all() if cond(song)]


class MaimaiPlates:
    _client: "MaimaiClient"

    _kind: str  # The kind of the plate, e.g. "将", "神".
    _version: str  # The version of the plate, e.g. "真", "舞".
    _versions: set[Version] = set()  # The matched versions set of the plate.
    _matched_songs: list[Song] = []
    _matched_scores: list[ScoreExtend] = []

    def __init__(self, client: "MaimaiClient") -> None:
        """@private"""
        self._client = client

    async def _configure(self, plate: str, scores: list[Score]) -> "MaimaiPlates":
        maimai_songs = await self._client.songs()
        self._version = plate_aliases.get(plate[0], plate[0])
        self._kind = plate_aliases.get(plate[1:], plate[1:])

        versions = list()  # in case of invalid plate, we will raise an error
        if self._version == "真":
            versions = [plate_to_version["初"], plate_to_version["真"]]
        if self._version in ["霸", "舞"]:
            versions = [ver for ver in plate_to_version.values() if ver.value < 20000]
        if plate_to_version.get(self._version):
            versions = [plate_to_version[self._version]]
        if not versions or self._kind not in ["将", "者", "极", "舞舞", "神"]:
            raise InvalidPlateError(f"Invalid plate: {self._version}{self._kind}")
        versions.append([ver for ver in plate_to_version.values() if ver.value > versions[-1].value][0])
        self._versions = set(versions)

        song_diff_versions: dict[str, int] = await self._client._cache.get("versions", namespace="songs") or {}
        versioned_matched_songs = set()
        for k, v in song_diff_versions.items():
            if any(v >= o.value and v < versions[i + 1].value for i, o in enumerate(versions[:-1])):
                versioned_matched_songs.add(int(k.split(" ")[0]))
        self._matched_songs = await self._client._multi_get(list(versioned_matched_songs), namespace="songs")

        versioned_joined_scores: dict[str, Score] = {}
        for score in scores:
            score_key = f"{score.id} {score.type} {score.level_index}"
            if score_version := song_diff_versions.get(score_key, None):
                if any(score_version >= o.value and score_version < versions[i + 1].value for i, o in enumerate(versions[:-1])):
                    if not (score.level_index == LevelIndex.ReMASTER and self.no_remaster):
                        versioned_joined_scores[score_key] = score._join(versioned_joined_scores.get(score_key, None))
        self._matched_scores = await MaimaiScores._get_extended(versioned_joined_scores.values(), maimai_songs)

        return self

    @cached_property
    def _major_type(self) -> SongType:
        return SongType.DX if any(ver.value > 20000 for ver in self._versions) else SongType.STANDARD

    @cached_property
    def no_remaster(self) -> bool:
        """Whether it is required to play ReMASTER levels in the plate.

        Only 舞 and 霸 plates require ReMASTER levels, others don't.
        """
        return self._version not in ["舞", "霸"]

    def _get_levels(self, song: Song) -> set[LevelIndex]:
        levels = set(diff.level_index for diff in song.get_difficulties(self._major_type))
        if self.no_remaster and LevelIndex.ReMASTER in levels:
            levels.remove(LevelIndex.ReMASTER)
        return levels

    async def get_remained(self) -> list[PlateObject]:
        """Get the remained songs and scores of the player on this plate.

        If player has ramained levels on one song, the song and ramained `level_index` will be included in the result, otherwise it won't.

        The distinct scores which NOT met the plate requirement will be included in the result, the finished scores won't.

        Returns:
            A list of `PlateObject` containing the song and the scores.
        """
        # Group scores by song ID to pre-fill the PlateObject.
        grouped = defaultdict(list)
        [grouped[score.id].append(score) for score in self._matched_scores]
        # Create PlateObject for each song with its levels and scores.
        results = {song.id: PlateObject(song=song, levels=self._get_levels(song), scores=grouped.get(song.id, [])) for song in self._matched_songs}

        def extract(score: ScoreExtend) -> None:
            results[score.id].scores.remove(score)
            if score.level_index in results[score.id].levels:
                results[score.id].levels.remove(score.level_index)

        if self._kind == "者":
            [extract(score) for score in self._matched_scores if score.rate.value <= RateType.A.value]
        elif self._kind == "将":
            [extract(score) for score in self._matched_scores if score.rate.value <= RateType.SSS.value]
        elif self._kind == "极":
            [extract(score) for score in self._matched_scores if score.fc and score.fc.value <= FCType.FC.value]
        elif self._kind == "舞舞":
            [extract(score) for score in self._matched_scores if score.fs and score.fs.value <= FSType.FSD.value]
        elif self._kind == "神":
            [extract(score) for score in self._matched_scores if score.fc and score.fc.value <= FCType.AP.value]

        return [plate for plate in results.values() if len(plate.levels) > 0]

    async def get_cleared(self) -> list[PlateObject]:
        """Get the cleared songs and scores of the player on this plate.

        If player has levels (one or more) that met the requirement on the song, the song and cleared `level_index` will be included in the result, otherwise it won't.

        The distinct scores which met the plate requirement will be included in the result, the unfinished scores won't.

        Returns:
            A list of `PlateObject` containing the song and the scores.
        """
        results = {song.id: PlateObject(song=song, levels=set(), scores=[]) for song in self._matched_songs}

        def insert(score: ScoreExtend) -> None:
            results[score.id].scores.append(score)
            results[score.id].levels.add(score.level_index)

        if self._kind == "者":
            [insert(score) for score in self._matched_scores if score.rate.value <= RateType.A.value]
        elif self._kind == "将":
            [insert(score) for score in self._matched_scores if score.rate.value <= RateType.SSS.value]
        elif self._kind == "极":
            [insert(score) for score in self._matched_scores if score.fc and score.fc.value <= FCType.FC.value]
        elif self._kind == "舞舞":
            [insert(score) for score in self._matched_scores if score.fs and score.fs.value <= FSType.FSD.value]
        elif self._kind == "神":
            [insert(score) for score in self._matched_scores if score.fc and score.fc.value <= FCType.AP.value]

        return [plate for plate in results.values() if len(plate.levels) > 0]

    async def get_played(self) -> list[PlateObject]:
        """Get the played songs and scores of the player on this plate.

        If player has ever played levels on the song, whether they met or not, the song and played `level_index` will be included in the result.

        All distinct scores will be included in the result.

        Returns:
            A list of `PlateObject` containing the song and the scores.
        """
        results = {song.id: PlateObject(song=song, levels=set(), scores=[]) for song in self._matched_songs}

        for score in self._matched_scores:
            results[score.id].scores.append(score)
            results[score.id].levels.add(score.level_index)

        return [plate for plate in results.values() if len(plate.levels) > 0]

    async def get_all(self) -> list[PlateObject]:
        """Get all songs and scores on this plate, usually used for overall statistics of the plate.

        All songs will be included in the result, with played `level_index`, whether they met or not.

        All distinct scores will be included in the result.

        Returns:
            A list of `PlateObject` containing the song and the scores.
        """
        results = {song.id: PlateObject(song=song, levels=set(), scores=[]) for song in self._matched_songs}

        for score in self._matched_scores:
            results[score.id].scores.append(score)
            results[score.id].levels.add(score.level_index)

        return [plate for plate in results.values()]

    async def count_played(self) -> int:
        """Get the number of played levels on this plate.

        Returns:
            The number of played levels on this plate.
        """
        return len([level for plate in await self.get_played() for level in plate.levels])

    async def count_cleared(self) -> int:
        """Get the number of cleared levels on this plate.

        Returns:
            The number of cleared levels on this plate.
        """
        return len([level for plate in await self.get_cleared() for level in plate.levels])

    async def count_remained(self) -> int:
        """Get the number of remained levels on this plate.

        Returns:
            The number of remained levels on this plate.
        """
        return len([level for plate in await self.get_remained() for level in plate.levels])

    async def count_all(self) -> int:
        """Get the number of all levels on this plate.

        Returns:
            The number of all levels on this plate.
        """
        return sum(len(self._get_levels(plate.song)) for plate in await self.get_all())


class MaimaiScores:
    _client: "MaimaiClient"

    scores: list[ScoreExtend]
    """All scores of the player."""
    scores_b35: list[ScoreExtend]
    """The b35 scores of the player."""
    scores_b15: list[ScoreExtend]
    """The b15 scores of the player."""
    rating: int
    """The total rating of the player."""
    rating_b35: int
    """The b35 rating of the player."""
    rating_b15: int
    """The b15 rating of the player."""

    def __init__(self, client: "MaimaiClient"):
        self._client = client

    async def configure(self, scores: list[Score], b50_only: bool = False) -> "MaimaiScores":
        """Initialize the scores by the scores list.

        This method will sort the scores by their dx_rating, dx_score and achievements, and split them into b35 and b15 scores.

        Args:
            scores: the scores list to initialize.
        Returns:
            The MaimaiScores object with the scores initialized.
        """
        maimai_songs = await self._client.songs()  # Ensure songs are configured.
        song_diff_versions: dict[str, int] = await self._client._cache.get("versions", namespace="songs") or {}
        self.scores, self.scores_b35, self.scores_b15 = [], [], []

        # Remove duplicates from scores based on id, type and level_index.
        scores_unique: dict[str, Score] = {}
        for score in scores:
            score_key = f"{score.id} {score.type} {score.level_index}"
            scores_unique[score_key] = score._compare(scores_unique.get(score_key, None))

        # Extend scores and categorize them into b35 and b15 based on their versions.
        self.scores = await MaimaiScores._get_extended(scores_unique.values(), maimai_songs)
        for score in self.scores:
            if score_version := song_diff_versions.get(f"{score.id} {score.type} {score.level_index}", None):
                (self.scores_b15 if score_version >= current_version.value else self.scores_b35).append(score)

        # Sort scores by dx_rating, dx_score and achievements, and limit the number of scores.
        self.scores_b35.sort(key=lambda score: (score.dx_rating or 0, score.dx_score or 0, score.achievements or 0), reverse=True)
        self.scores_b15.sort(key=lambda score: (score.dx_rating or 0, score.dx_score or 0, score.achievements or 0), reverse=True)
        self.scores_b35 = self.scores_b35[:35]
        self.scores_b15 = self.scores_b15[:15]
        self.scores = self.scores_b35 + self.scores_b15 if b50_only else self.scores

        # Calculate the total rating.
        self.rating_b35 = int(sum((score.dx_rating or 0) for score in self.scores_b35))
        self.rating_b15 = int(sum((score.dx_rating or 0) for score in self.scores_b15))
        self.rating = self.rating_b35 + self.rating_b15

        return self

    @staticmethod
    async def _get_mapping(scores: Iterable[T], maimai_songs: MaimaiSongs) -> AsyncGenerator[tuple[Song, SongDifficulty, T], None]:
        required_songs = await maimai_songs.get_batch(set(score.id for score in scores))
        required_songs_dict = {song.id: song for song in required_songs if song is not None}
        for score in scores:
            song = required_songs_dict.get(score.id, None)
            diff = song.get_difficulty(score.type, score.level_index) if song else None
            if score and song and diff:
                yield (song, diff, score)

    @staticmethod
    async def _get_extended(scores: Iterable[Score], maimai_songs: MaimaiSongs) -> list[ScoreExtend]:
        extended_scores = []
        async for song, diff, score in MaimaiScores._get_mapping(scores, maimai_songs):
            extended_score = ScoreExtend(
                **dataclasses.asdict(score),
                title=song.title,
                level_value=diff.level_value,
                level_dx_score=(diff.tap_num + diff.hold_num + diff.slide_num + diff.break_num + diff.touch_num) * 3,
            )
            extended_score.level = diff.level  # Ensure level is set correctly.
            extended_scores.append(extended_score)
        return extended_scores

    async def get_mapping(self) -> list[tuple[Song, SongDifficulty, ScoreExtend]]:
        """Get all scores with their corresponding songs.

        This method will return a list of tuples, each containing a song, its corresponding difficulty, and the score.

        If the song or difficulty is not found, the whole tuple will be excluded from the result.

        Args:
            override_scores: a list of scores to override the current instance scores, defaults to UNSET.
        Returns:
            A list of tuples, each containing (song, difficulty, score).
        """
        maimai_songs, result = await self._client.songs(), []
        async for v in self._get_mapping(self.scores, maimai_songs):
            result.append(v)
        return result

    def get_player_bests(self) -> PlayerBests:
        """Get the best scores of the player.

        This method will return a PlayerBests object containing the best scores of the player, sorted by their dx_rating, dx_score and achievements.

        Returns:
            A PlayerBests object containing the best scores of the player.
        """
        return PlayerBests(
            rating=self.rating,
            rating_b35=self.rating_b35,
            rating_b15=self.rating_b15,
            scores_b35=self.scores_b35,
            scores_b15=self.scores_b15,
        )

    def by_song(
        self, song_id: int, song_type: Union[SongType, _UnsetSentinel] = UNSET, level_index: Union[LevelIndex, _UnsetSentinel] = UNSET
    ) -> list[ScoreExtend]:
        """Get scores of the song on that type and level_index.

        If song_type or level_index is not provided, it won't be filtered by that attribute.

        Args:
            song_id: the ID of the song to get the scores by.
            song_type: the type of the song to get the scores by, defaults to None.
            level_index: the level index of the song to get the scores by, defaults to None.
        Returns:
            A list of scores that match the song ID, type and level index.
            If no score is found, an empty list will be returned.
        """
        return [
            score
            for score in self.scores
            if score.id == song_id
            and (score.type == song_type or isinstance(song_type, _UnsetSentinel))
            and (score.level_index == level_index or isinstance(level_index, _UnsetSentinel))
        ]

    def filter(self, **kwargs) -> list[Score]:
        """Filter scores by their attributes.

        Make sure the attribute is of the score, and the value is of the same type. All conditions are connected by AND.

        Args:
            kwargs: the attributes to filter the scores by.
        Returns:
            an iterator of scores that match all the conditions, yields no items if no score is found.
        """
        return [score for score in self.scores if all(getattr(score, key) == value for key, value in kwargs.items() if value is not None)]


class MaimaiAreas:
    _client: "MaimaiClient"
    _lang: str

    def __init__(self, client: "MaimaiClient") -> None:
        """@private"""
        self._client = client

    async def _configure(self, lang: str, provider: Union[IAreaProvider, _UnsetSentinel]) -> "MaimaiAreas":
        self._lang = lang
        cache_obj, cache_ttl = self._client._cache, self._client._cache_ttl
        # Check if the provider is unset, which means we want to access the cache directly.
        if isinstance(provider, _UnsetSentinel):
            if await self._client._cache.get("provider", None, namespace=f"areas_{lang}") is not None:
                return self
        # Really assign the unset provider to the default one.
        provider = provider if not isinstance(provider, _UnsetSentinel) else LocalProvider()
        # Check if the current provider hash is different from the previous one, which means we need to reconfigure.
        current_provider_hash = provider._hash()
        previous_provider_hash = await cache_obj.get("provider", "", namespace=f"areas_{lang}")
        if current_provider_hash != previous_provider_hash:
            areas = await provider.get_areas(lang, self._client)
            await asyncio.gather(
                cache_obj.set("provider", hash(provider), ttl=cache_ttl, namespace=f"areas_{lang}"),  # provider
                cache_obj.set("ids", [area.id for area in areas.values()], namespace=f"areas_{lang}"),  # ids
                cache_obj.multi_set(iter((k, v) for k, v in areas.items()), namespace=f"areas_{lang}"),  # areas
            )
        return self

    async def get_all(self) -> list[Area]:
        """All areas as list.

        This method will iterate all areas in the cache. Unless you really need to iterate all areas, you should use `by_id` or `by_name` instead.

        Returns:
            A list of all areas in the cache, return an empty list if no area is found.
        """
        area_ids: Optional[list[int]] = await self._client._cache.get("ids", namespace=f"areas_{self._lang}")
        assert area_ids is not None, "Areas not found in cache, please call configure() first."
        return await self._client._multi_get(area_ids, namespace=f"areas_{self._lang}")

    async def get_batch(self, ids: Iterable[str]) -> list[Area]:
        """Get areas by their IDs.

        Args:
            ids: the IDs of the areas.
        Returns:
            A list of areas if they exist, otherwise return an empty list.
        """
        return await self._client._multi_get(ids, namespace=f"areas_{self._lang}")

    async def by_id(self, id: str) -> Optional[Area]:
        """Get an area by its ID.

        Args:
            id: the ID of the area.
        Returns:
            the area if it exists, otherwise return None.
        """
        return await self._client._cache.get(id, namespace=f"areas_{self._lang}")

    async def by_name(self, name: str) -> Optional[Area]:
        """Get an area by its name, language-sensitive.

        Args:
            name: the name of the area.
        Returns:
            the area if it exists, otherwise return None.
        """
        return next((area for area in await self.get_all() if area.name == name), None)


class MaimaiClient:
    """The main client of maimai.py."""

    _client: AsyncClient
    _cache: BaseCache
    _cache_ttl: int

    def __new__(cls, *args, **kwargs):
        if hasattr(cls, "_instance"):
            warn_message = (
                "MaimaiClient is a singleton, args are ignored in this case, due to the singleton nature. "
                "If you think this is a mistake, please check MaimaiClientMultithreading. "
            )
            warnings.warn(warn_message, stacklevel=2)
            return cls._instance
        orig = super(MaimaiClient, cls)
        cls._instance = orig.__new__(cls)
        return cls._instance

    def __init__(
        self,
        timeout: float = 20.0,
        cache: Union[BaseCache, _UnsetSentinel] = UNSET,
        cache_ttl: int = 60 * 60 * 24,
        **kwargs,
    ) -> None:
        """Initialize the maimai.py client.

        Args:
            timeout: the timeout of the requests, defaults to 20.0.
            cache: the cache to use, defaults to `aiocache.SimpleMemoryCache()`.
            cache_ttl: the TTL of the cache, defaults to 60 * 60 * 24.
            kwargs: other arguments to pass to the `httpx.AsyncClient`.
        """
        self._client = AsyncClient(timeout=timeout, **kwargs)
        self._cache = SimpleMemoryCache() if isinstance(cache, _UnsetSentinel) else cache
        self._cache_ttl = cache_ttl

    async def _multi_get(self, keys: Iterable[Any], namespace: Optional[str] = None) -> list[Any]:
        keys_list = list(keys)
        if len(keys_list) != 0:
            return await self._cache.multi_get(keys_list, namespace=namespace)
        return []

    async def songs(
        self,
        provider: Union[ISongProvider, _UnsetSentinel] = UNSET,
        alias_provider: Union[IAliasProvider, None, _UnsetSentinel] = UNSET,
        curve_provider: Union[ICurveProvider, None, _UnsetSentinel] = UNSET,
    ) -> MaimaiSongs:
        """Fetch all maimai songs from the provider.

        Available providers: `DivingFishProvider`, `LXNSProvider`.

        Available alias providers: `YuzuProvider`, `LXNSProvider`.

        Available curve providers: `DivingFishProvider`.

        Args:
            provider: override the data source to fetch the player from, defaults to `LXNSProvider`.
            alias_provider: override the data source to fetch the song aliases from, defaults to `YuzuProvider`.
            curve_provider: override the data source to fetch the song curves from, defaults to `None`.
        Returns:
            A wrapper of the song list, for easier access and filtering.
        Raises:
            httpx.HTTPError: Request failed due to network issues.
        """
        songs = MaimaiSongs(self)
        return await songs._configure(provider, alias_provider, curve_provider)

    async def players(
        self,
        identifier: PlayerIdentifier,
        provider: IPlayerProvider = LXNSProvider(),
    ) -> Player:
        """Fetch player data from the provider.

        Available providers: `DivingFishProvider`, `LXNSProvider`, `ArcadeProvider`.

        Possible returns: `DivingFishPlayer`, `LXNSPlayer`, `ArcadePlayer`.

        Args:
            identifier: the identifier of the player to fetch, e.g. `PlayerIdentifier(username="turou")`.
            provider: the data source to fetch the player from, defaults to `LXNSProvider`.
        Returns:
            The player object of the player, with all the data fetched. Depending on the provider, it may contain different objects that derived from `Player`.
        Raises:
            InvalidPlayerIdentifierError: Player identifier is invalid for the provider, or player is not found.
            InvalidDeveloperTokenError: Developer token is not provided or token is invalid.
            PrivacyLimitationError: The user has not accepted the 3rd party to access the data.
            httpx.HTTPError: Request failed due to network issues.
        Raises:
            TitleServerNetworkError: Only for ArcadeProvider, maimai title server related errors, possibly network problems.
            TitleServerBlockedError: Only for ArcadeProvider, maimai title server blocked the request, possibly due to ip filtered.
            ArcadeIdentifierError: Only for ArcadeProvider, maimai user id is invalid, or the user is not found.
        """
        return await provider.get_player(identifier, self)

    async def scores(
        self,
        identifier: PlayerIdentifier,
        provider: IScoreProvider = LXNSProvider(),
    ) -> MaimaiScores:
        """Fetch player's ALL scores from the provider.

        All scores of the player will be fetched, if you want to fetch only the best scores (for better performance), use `maimai.bests()` instead.

        For WechatProvider, PlayerIdentifier must have the `credentials` attribute, we suggest you to use the `maimai.wechat()` method to get the identifier.
        Also, PlayerIdentifier should not be cached or stored in the database, as the cookies may expire at any time.

        For ArcadeProvider, PlayerIdentifier must have the `credentials` attribute, which is the player's encrypted userId, can be detrived from `maimai.qrcode()`.
        Credentials can be reused, since it won't expire, also, userId is encrypted, can't be used in any other cases outside the maimai.py

        For more information about the PlayerIdentifier of providers, please refer to the documentation of each provider.

        Available providers: `DivingFishProvider`, `LXNSProvider`, `WechatProvider`, `ArcadeProvider`.

        Args:
            identifier: the identifier of the player to fetch, e.g. `PlayerIdentifier(friend_code=664994421382429)`.
            provider: the data source to fetch the player and scores from, defaults to `LXNSProvider`.
        Returns:
            The scores object of the player, with all the data fetched.
        Raises:
            InvalidPlayerIdentifierError: Player identifier is invalid for the provider, or player is not found.
            InvalidDeveloperTokenError: Developer token is not provided or token is invalid.
            PrivacyLimitationError: The user has not accepted the 3rd party to access the data.
            httpx.HTTPError: Request failed due to network issues.
        Raises:
            TitleServerNetworkError: Only for ArcadeProvider, maimai title server related errors, possibly network problems.
            TitleServerBlockedError: Only for ArcadeProvider, maimai title server blocked the request, possibly due to ip filtered.
            ArcadeIdentifierError: Only for ArcadeProvider, maimai user id is invalid, or the user is not found.
        """
        scores = await provider.get_scores_all(identifier, self)

        maimai_scores = MaimaiScores(self)
        return await maimai_scores.configure(scores)

    async def bests(
        self,
        identifier: PlayerIdentifier,
        provider: IScoreProvider = LXNSProvider(),
    ) -> MaimaiScores:
        """Fetch player's B50 scores from the provider.

        Though MaimaiScores is used, this method will only return the best 50 scores. if you want all scores, please use `maimai.scores()` method instead.

        For WechatProvider, PlayerIdentifier must have the `credentials` attribute, we suggest you to use the `maimai.wechat()` method to get the identifier.
        Also, PlayerIdentifier should not be cached or stored in the database, as the cookies may expire at any time.

        For ArcadeProvider, PlayerIdentifier must have the `credentials` attribute, which is the player's encrypted userId, can be detrived from `maimai.qrcode()`.
        Credentials can be reused, since it won't expire, also, userId is encrypted, can't be used in any other cases outside the maimai.py

        For more information about the PlayerIdentifier of providers, please refer to the documentation of each provider.

        Available providers: `DivingFishProvider`, `LXNSProvider`, `WechatProvider`, `ArcadeProvider`.

        Args:
            identifier: the identifier of the player to fetch, e.g. `PlayerIdentifier(friend_code=664994421382429)`.
            provider: the data source to fetch the player and scores from, defaults to `LXNSProvider`.
        Returns:
            The scores object of the player, with all the data fetched.
        Raises:
            InvalidPlayerIdentifierError: Player identifier is invalid for the provider, or player is not found.
            InvalidDeveloperTokenError: Developer token is not provided or token is invalid.
            PrivacyLimitationError: The user has not accepted the 3rd party to access the data.
            httpx.HTTPError: Request failed due to network issues.
        Raises:
            TitleServerNetworkError: Only for ArcadeProvider, maimai title server related errors, possibly network problems.
            TitleServerBlockedError: Only for ArcadeProvider, maimai title server blocked the request, possibly due to ip filtered.
            ArcadeIdentifierError: Only for ArcadeProvider, maimai user id is invalid, or the user is not found.
        """
        maimai_scores = MaimaiScores(self)
        best_scores = await provider.get_scores_best(identifier, self)
        return await maimai_scores.configure(best_scores, b50_only=True)

    async def minfo(
        self,
        song: Union[Song, int, str],
        identifier: PlayerIdentifier,
        provider: IScoreProvider = LXNSProvider(),
    ) -> Optional[PlayerSong]:
        """Fetch player's scores on the specific song.

        This method will return all scores of the player on the song.

        For more information about the PlayerIdentifier of providers, please refer to the documentation of each provider.

        Available providers: `DivingFishProvider`, `LXNSProvider`, `WechatProvider`, `ArcadeProvider`.

        Args:
            song: the song to fetch the scores from, can be a `Song` object, or a song_id (int), or keywords (str).
            identifier: the identifier of the player to fetch, e.g. `PlayerIdentifier(friend_code=664994421382429)`.
            provider: the data source to fetch the player and scores from, defaults to `LXNSProvider`.
        Returns:
            A tuple of the song and the scores of the player on the song. If the song is not found, the song will be None and scores will be an empty list.
            If the song is found, the song will be a `Song` object and scores will be a list of `Score` objects.
        Raises:
            InvalidPlayerIdentifierError: Player identifier is invalid for the provider, or player is not found.
            InvalidDeveloperTokenError: Developer token is not provided or token is invalid.
            PrivacyLimitationError: The user has not accepted the 3rd party to access the data.
            httpx.HTTPError: Request failed due to network issues.
        Raises:
            TitleServerNetworkError: Only for ArcadeProvider, maimai title server related errors, possibly network problems.
            TitleServerBlockedError: Only for ArcadeProvider, maimai title server blocked the request, possibly due to ip filtered.
            ArcadeIdentifierError: Only for ArcadeProvider, maimai user id is invalid, or the user is not found.
        """
        maimai_songs, scores = await self.songs(), []
        if isinstance(song, str):
            search_result = await maimai_songs.by_keywords(song)
            song = search_result[0] if len(search_result) > 0 else song
        if isinstance(song, int):
            search_result = await maimai_songs.by_id(song)
            song = search_result if search_result is not None else song
        if isinstance(song, Song):
            scores = await provider.get_scores_one(identifier, song, self)
            extended_scores = await MaimaiScores._get_extended(scores, maimai_songs)
            return PlayerSong(song, extended_scores)

    async def regions(self, identifier: PlayerIdentifier, provider: IRegionProvider = ArcadeProvider()) -> list[PlayerRegion]:
        """Get the player's regions that they have played.

        Args:
            identifier: the identifier of the player to fetch, e.g. `PlayerIdentifier(credentials="encrypted_user_id")`.
            provider: the data source to fetch the player from, defaults to `ArcadeProvider`.
        Returns:
            The list of regions that the player has played.
        Raises:
            TitleServerNetworkError: Only for ArcadeProvider, maimai title server related errors, possibly network problems.
            TitleServerBlockedError: Only for ArcadeProvider, maimai title server blocked the request, possibly due to ip filtered.
            ArcadeIdentifierError: Only for ArcadeProvider, maimai user id is invalid, or the user is not found.
        """
        return await provider.get_regions(identifier, self)

    async def updates(
        self,
        identifier: PlayerIdentifier,
        scores: Iterable[Score],
        provider: IScoreUpdateProvider = LXNSProvider(),
    ) -> None:
        """Update player's scores to the provider.

        This method is used to update the player's scores to the provider, usually used for updating scores fetched from other providers.

        For more information about the PlayerIdentifier of providers, please refer to the documentation of each provider.

        Available providers: `DivingFishProvider`, `LXNSProvider`.

        Args:
            identifier: the identifier of the player to update, e.g. `PlayerIdentifier(friend_code=664994421382429)`.
            scores: the scores to update, usually the scores fetched from other providers.
            provider: the data source to update the player scores to, defaults to `LXNSProvider`.
        Returns:
            Nothing, failures will raise exceptions.
        Raises:
            InvalidPlayerIdentifierError: Player identifier is invalid for the provider, or player is not found, or the import token / password is invalid.
            InvalidDeveloperTokenError: Developer token is not provided or token is invalid.
            PrivacyLimitationError: The user has not accepted the 3rd party to access the data.
            httpx.HTTPError: Request failed due to network issues.
        """
        await provider.update_scores(identifier, scores, self)

    async def plates(
        self,
        identifier: PlayerIdentifier,
        plate: str,
        provider: IScoreProvider = LXNSProvider(),
    ) -> MaimaiPlates:
        """Get the plate achievement of the given player and plate.

        Available providers: `DivingFishProvider`, `LXNSProvider`, `ArcadeProvider`.

        Args:
            identifier: the identifier of the player to fetch, e.g. `PlayerIdentifier(friend_code=664994421382429)`.
            plate: the name of the plate, e.g. "樱将", "真舞舞".
            provider: the data source to fetch the player and scores from, defaults to `LXNSProvider`.
        Returns:
            A wrapper of the plate achievement, with plate information, and matched player scores.
        Raises:
            InvalidPlayerIdentifierError: Player identifier is invalid for the provider, or player is not found.
            InvalidPlateError: Provided version or plate is invalid.
            InvalidDeveloperTokenError: Developer token is not provided or token is invalid.
            PrivacyLimitationError: The user has not accepted the 3rd party to access the data.
            httpx.HTTPError: Request failed due to network issues.
        Raises:
            TitleServerNetworkError: Only for ArcadeProvider, maimai title server related errors, possibly network problems.
            TitleServerBlockedError: Only for ArcadeProvider, maimai title server blocked the request, possibly due to ip filtered.
            ArcadeIdentifierError: Only for ArcadeProvider, maimai user id is invalid, or the user is not found.
        """
        scores = await provider.get_scores_all(identifier, self)
        maimai_plates = MaimaiPlates(self)
        return await maimai_plates._configure(plate, scores)

    async def identifiers(
        self,
        code: Union[str, dict[str, str]],
        provider: Union[IPlayerIdentifierProvider, _UnsetSentinel] = UNSET,
    ) -> PlayerIdentifier:
        """Get the player identifier from the provider.

        This method is combination of `maimai.wechat()` and `maimai.qrcode()`, which will return the player identifier of the player.

        For WechatProvider, code should be a dictionary with `r`, `t`, `code`, and `state` keys, or a string that contains the URL parameters.

        For ArcadeProvider, code should be a string that begins with `SGWCMAID`, which is the QR code of the player.

        Available providers: `WechatProvider`, `ArcadeProvider`.

        Args:
            code: the code to get the player identifier, can be a string or a dictionary with `r`, `t`, `code`, and `state` keys.
            provider: override the default provider, defaults to `ArcadeProvider`.
        Returns:
            The player identifier of the player.
        Raises:
            InvalidWechatTokenError: Wechat token is expired, please re-authorize.
            AimeServerError: Maimai Aime server error, may be invalid QR code or QR code has expired.
            httpx.HTTPError: Request failed due to network issues.
        """
        if isinstance(provider, _UnsetSentinel):
            provider = ArcadeProvider()
        return await provider.get_identifier(code, self)

    async def items(self, item: Type[PlayerItemType], provider: Union[IItemListProvider, _UnsetSentinel] = UNSET) -> MaimaiItems[PlayerItemType]:
        """Fetch maimai player items from the cache default provider.

        Available items: `PlayerIcon`, `PlayerNamePlate`, `PlayerFrame`, `PlayerTrophy`, `PlayerChara`, `PlayerPartner`.

        Args:
            item: the item type to fetch, e.g. `PlayerIcon`.
            provider: override the default item list provider, defaults to `LXNSProvider` and `LocalProvider`.
        Returns:
            A wrapper of the item list, for easier access and filtering.
        Raises:
            FileNotFoundError: The item file is not found.
            httpx.HTTPError: Request failed due to network issues.
        """
        maimai_items = MaimaiItems[PlayerItemType](self, item._namespace())
        return await maimai_items._configure(provider)

    async def areas(self, lang: Literal["ja", "zh"] = "ja", provider: IAreaProvider = LocalProvider()) -> MaimaiAreas:
        """Fetch maimai areas from the provider.

        Available providers: `LocalProvider`.

        Args:
            lang: the language of the area to fetch, available languages: `ja`, `zh`.
            provider: override the default area provider, defaults to `ArcadeProvider`.
        Returns:
            A wrapper of the area list, for easier access and filtering.
        Raises:
            FileNotFoundError: The area file is not found.
        """
        maimai_areas = MaimaiAreas(self)
        return await maimai_areas._configure(lang, provider)

    async def wechat(
        self,
        r: Optional[str] = None,
        t: Optional[str] = None,
        code: Optional[str] = None,
        state: Optional[str] = None,
    ) -> Union[str, PlayerIdentifier]:
        """Get the player identifier from the Wahlap Wechat OffiAccount.

        Call the method with no parameters to get the URL, then redirect the user to the URL with your mitmproxy enabled.

        Your mitmproxy should intercept the response from tgk-wcaime.wahlap.com, then call the method with the parameters from the intercepted response.

        With the parameters from specific user's response, the method will return the user's player identifier.

        Never cache or store the player identifier, as the cookies may expire at any time.

        Args:
            r: the r parameter from the request, defaults to None.
            t: the t parameter from the request, defaults to None.
            code: the code parameter from the request, defaults to None.
            state: the state parameter from the request, defaults to None.
        Returns:
            The player identifier if all parameters are provided, otherwise return the URL to get the identifier.
        Raises:
            WechatTokenExpiredError: Wechat token is expired, please re-authorize.
            httpx.HTTPError: Request failed due to network issues.
        """
        if r is None or t is None or code is None or state is None:
            resp = await self._client.get("https://tgk-wcaime.wahlap.com/wc_auth/oauth/authorize/maimai-dx")
            return resp.headers["location"].replace("redirect_uri=https", "redirect_uri=http")
        return await WechatProvider().get_identifier({"r": r, "t": t, "code": code, "state": state}, self)

    async def qrcode(self, qrcode: str, http_proxy: Optional[str] = None) -> PlayerIdentifier:
        """Get the player identifier from the Wahlap QR code.

        Player identifier is the encrypted userId, can't be used in any other cases outside the maimai.py.

        Args:
            qrcode: the QR code of the player, should begin with SGWCMAID.
            http_proxy: the http proxy to use for the request, defaults to None.
        Returns:
            The player identifier of the player.
        Raises:
            AimeServerError: Maimai Aime server error, may be invalid QR code or QR code has expired.
        """
        provider = ArcadeProvider(http_proxy=http_proxy)
        return await provider.get_identifier(qrcode, self)


class MaimaiClientMultithreading(MaimaiClient):
    """Multi-threading version of maimai.py.
    Introduced by issue #28. Users who want to share the same client instance across multiple threads can use this class.
    """

    def __new__(cls, *args, **kwargs):
        # Override the singleton behavior by always creating a new instance
        return super(MaimaiClient, cls).__new__(cls)



import hashlib
from typing import TYPE_CHECKING

from maimai_py.models import *
from maimai_py.providers.local import LocalProvider

from .base import IAreaProvider, IItemListProvider, ISongProvider
from .divingfish import DivingFishProvider
from .lxns import LXNSProvider

if TYPE_CHECKING:
    from maimai_py.maimai import MaimaiClient


class HybridProvider(ISongProvider, IItemListProvider, IAreaProvider):
    """The provider that fetches data from the LXNS and local, and hybrids them together.

    This provider is used to api endpoints creation, which are not normally used by the client.
    """

    base_url_lxns = LXNSProvider.base_url
    base_url_divingfish = DivingFishProvider.base_url
    provider_local = LocalProvider()
    provider_lxns = LXNSProvider()

    def _hash(self) -> str:
        return hashlib.md5(b"hybrid").hexdigest()

    async def get_songs(self, client: "MaimaiClient") -> list[Song]:
        return await self.provider_lxns.get_songs(client)

    async def get_icons(self, client: "MaimaiClient") -> dict[int, PlayerIcon]:
        return await self.provider_lxns.get_icons(client)

    async def get_nameplates(self, client: "MaimaiClient") -> dict[int, PlayerNamePlate]:
        return await self.provider_lxns.get_nameplates(client)

    async def get_frames(self, client: "MaimaiClient") -> dict[int, PlayerFrame]:
        return await self.provider_lxns.get_frames(client)

    async def get_partners(self, client: "MaimaiClient") -> dict[int, PlayerPartner]:
        return await self.provider_local.get_partners(client)

    async def get_charas(self, client: "MaimaiClient") -> dict[int, PlayerChara]:
        return await self.provider_local.get_charas(client)

    async def get_trophies(self, client: "MaimaiClient") -> dict[int, PlayerTrophy]:
        return await self.provider_local.get_trophies(client)

    async def get_areas(self, lang: str, client: "MaimaiClient") -> dict[str, Area]:
        return await self.provider_local.get_areas(lang, client)



from abc import abstractmethod
from typing import TYPE_CHECKING, Iterable

from maimai_py.models import *

if TYPE_CHECKING:
    from maimai_py.maimai import MaimaiClient


class IProvider:
    @abstractmethod
    def _hash(self) -> str: ...


class ISongProvider(IProvider):
    """The provider that fetches songs from a specific source.

    Available providers: `DivingFishProvider`, `LXNSProvider`
    """

    @abstractmethod
    async def get_songs(self, client: "MaimaiClient") -> list[Song]:
        """@private"""
        raise NotImplementedError()


class IAliasProvider(IProvider):
    """The provider that fetches song aliases from a specific source.

    Available providers: `YuzuProvider`, `LXNSProvider`
    """

    @abstractmethod
    async def get_aliases(self, client: "MaimaiClient") -> dict[int, list[str]]:
        """@private"""
        raise NotImplementedError()


class IPlayerProvider(IProvider):
    """The provider that fetches players from a specific source.

    Available providers: `DivingFishProvider`, `LXNSProvider`
    """

    @abstractmethod
    async def get_player(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> Player:
        """@private"""
        raise NotImplementedError()


class IScoreProvider(IProvider):
    """The provider that fetches scores from a specific source.

    Available providers: `DivingFishProvider`, `LXNSProvider`, `WechatProvider`
    """

    async def get_scores_one(self, identifier: PlayerIdentifier, song: Song, client: "MaimaiClient") -> list[Score]:
        """@private"""
        scores = await self.get_scores_all(identifier, client)
        return [score for score in scores if score.id == song.id]

    async def get_scores_best(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[Score]:
        """@private"""
        return await self.get_scores_all(identifier, client)

    @abstractmethod
    async def get_scores_all(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[Score]:
        """@private"""
        raise NotImplementedError()


class IScoreUpdateProvider(IProvider):
    """The provider that updates scores to a specific source.

    Available providers: `DivingFishProvider`, `LXNSProvider`
    """

    @abstractmethod
    async def update_scores(self, identifier: PlayerIdentifier, scores: Iterable[Score], client: "MaimaiClient") -> None:
        """@private"""
        raise NotImplementedError()


class ICurveProvider(IProvider):
    """The provider that fetches statistics curves from a specific source.

    Available providers: `DivingFishProvider`
    """

    @abstractmethod
    async def get_curves(self, client: "MaimaiClient") -> dict[tuple[int, SongType], list[CurveObject]]:
        """@private"""
        raise NotImplementedError()


class IRegionProvider(IProvider):
    """The provider that fetches player regions from a specific source.

    Available providers: `ArcadeProvider`
    """

    @abstractmethod
    async def get_regions(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[PlayerRegion]:
        """@private"""
        raise NotImplementedError()


class IItemListProvider(IProvider):
    """The provider that fetches player item list data from a specific source.

    Available providers: `LXNSProvider`, `LocalProvider`
    """

    @abstractmethod
    async def get_icons(self, client: "MaimaiClient") -> dict[int, PlayerIcon]:
        """@private"""
        raise NotImplementedError()

    @abstractmethod
    async def get_nameplates(self, client: "MaimaiClient") -> dict[int, PlayerNamePlate]:
        """@private"""
        raise NotImplementedError()

    @abstractmethod
    async def get_frames(self, client: "MaimaiClient") -> dict[int, PlayerFrame]:
        """@private"""
        raise NotImplementedError()

    @abstractmethod
    async def get_partners(self, client: "MaimaiClient") -> dict[int, PlayerPartner]:
        """@private"""
        raise NotImplementedError()

    @abstractmethod
    async def get_charas(self, client: "MaimaiClient") -> dict[int, PlayerChara]:
        """@private"""
        raise NotImplementedError()

    @abstractmethod
    async def get_trophies(self, client: "MaimaiClient") -> dict[int, PlayerTrophy]:
        """@private"""
        raise NotImplementedError()


class IAreaProvider(IProvider):
    """The provider that fetches area data from a specific source.

    Available providers: `LocalProvider`
    """

    @abstractmethod
    async def get_areas(self, lang: str, client: "MaimaiClient") -> dict[str, Area]:
        """@private"""
        raise NotImplementedError()


class IPlayerIdentifierProvider(IProvider):
    """The provider that fetches player identifiers from a specific source.

    Available providers: `ArcadeProvider`
    """

    @abstractmethod
    async def get_identifier(self, code: Union[str, dict[str, str]], client: "MaimaiClient") -> PlayerIdentifier:
        """@private"""
        raise NotImplementedError()



import re
from dataclasses import dataclass
from typing import Optional

from lxml import etree

link_dx_score = [372, 522, 942, 924, 1425]


@dataclass
class HTMLScore:
    __slots__ = ["title", "level", "level_index", "type", "achievements", "dx_score", "rate", "fc", "fs", "ds"]
    title: str
    level: str
    level_index: int
    type: str
    achievements: float
    dx_score: int
    rate: str
    fc: str
    fs: str
    ds: int


def get_data_from_div(div) -> Optional[HTMLScore]:
    form = div.find(".//form")
    if form is None:
        return None

    # Find img element and get src attribute
    img = form.find(".//img")
    if img is None:
        return None

    img_src = img.get("src", "")

    # Determine type (SD or DX)
    if not re.search(r"diff_(.*).png", img_src):
        matched = re.search(r"music_(.*).png", img_src)
        type_ = "SD" if matched and matched.group(1) == "standard" else "DX"
    elif form.getparent().getparent().get("id") is not None:
        parent_id = form.getparent().getparent().get("id", "")
        type_ = "SD" if parent_id[:3] == "sta" else "DX"
    else:
        next_sibling = form.getparent().getnext()
        if next_sibling is not None:
            src = next_sibling.get("src", "")
            matched = re.search(r"_(.*).png", src)
            type_ = "SD" if matched and matched.group(1) == "standard" else "DX"
        else:
            type_ = "DX"  # Default

    def get_level_index(src: str) -> int:
        if src.find("remaster") != -1:
            return 4
        elif src.find("master") != -1:
            return 3
        elif src.find("expert") != -1:
            return 2
        elif src.find("advanced") != -1:
            return 1
        elif src.find("basic") != -1:
            return 0
        else:
            return -1

    def get_music_icon(src: str) -> str:
        matched = re.search(r"music_icon_(.+?)\.png", src)
        return matched.group(1) if matched and matched.group(1) != "back" else ""

    def get_dx_score(element) -> tuple[int, int]:
        elem_text = "".join(element.itertext())

        parts = elem_text.strip().split("/")
        if len(parts) != 2:
            return (0, 0)

        try:
            score = int(parts[0].replace(" ", "").replace(",", ""))
            full_score = int(parts[1].replace(" ", "").replace(",", ""))
            return (score, full_score)
        except (ValueError, IndexError):
            return (0, 0)

    # Extract data from form elements
    try:
        title_elem = form.xpath(".//div[contains(@class, 'music_name_block')]")
        level_elem = form.xpath(".//div[contains(@class, 'music_lv_block')]")
        score_elem = form.xpath(".//div[contains(@class, 'music_score_block')]")

        title = title_elem[0].text if title_elem else ""
        if title != "\u3000":  # Corner case for id 1422 (如月车站)
            title = title.strip()
        level = level_elem[0].text.strip() if level_elem else ""
        level_index = get_level_index(img_src)

        if len(score_elem) != 0:
            achievements = float(score_elem[0].text.strip()[:-1]) if score_elem else 0.0
            dx_score, full_dx_score = get_dx_score(score_elem[1] if score_elem else None)

            # Find icon elements
            icon_elems = form.xpath(".//img[contains(@src, 'music_icon')]")
            fs = fc = rate = ""

            if len(icon_elems) >= 3:
                fs = get_music_icon(icon_elems[0].get("src", ""))
                fc = get_music_icon(icon_elems[1].get("src", ""))
                rate = get_music_icon(icon_elems[2].get("src", ""))

            if title == "Link" and full_dx_score != link_dx_score[level_index]:
                title = "Link(CoF)"

            return HTMLScore(
                title=title,
                level=level,
                level_index=level_index,
                type=type_,
                achievements=achievements,
                dx_score=dx_score,
                rate=rate,
                fc=fc,
                fs=fs,
                ds=0,
            )
    except (IndexError, AttributeError):
        return None


def wmdx_html2json(html: str) -> list[HTMLScore]:
    parser = etree.HTMLParser()
    root = etree.fromstring(html, parser)

    # Find all divs with class "w_450 m_15 p_r f_0"
    divs = root.xpath("//div[contains(@class, 'w_450') and contains(@class, 'm_15') and contains(@class, 'p_r') and contains(@class, 'f_0')]")

    results = []
    for div in divs:
        score = get_data_from_div(div)
        if score is not None:
            results.append(score)

    del parser, root, divs

    return results



from dataclasses import dataclass
from datetime import datetime
from typing import Any, MutableMapping, Optional, Sequence, Union

from maimai_py.enums import *
from maimai_py.exceptions import *
from maimai_py.utils import UNSET, _UnsetSentinel


@dataclass
class Song:
    __slots__ = ("id", "title", "artist", "genre", "bpm", "map", "version", "rights", "aliases", "disabled", "difficulties")

    id: int
    title: str
    artist: str
    genre: Genre
    bpm: int
    map: Optional[str]
    version: int
    rights: Optional[str]
    aliases: Optional[list[str]]
    disabled: bool
    difficulties: "SongDifficulties"

    def get_difficulty(self, type: SongType, level_index: LevelIndex) -> Optional["SongDifficulty"]:
        """Get the exact difficulty of this song by type and level index.

        Args:
            type: The type of the song (DX, STANDARD, UTAGE).
            level_index: The level index of the difficulty.
        Returns:
            The difficulty object if found, otherwise None.
        """
        if type == SongType.DX:
            return next((diff for diff in self.difficulties.dx if diff.level_index == level_index), None)
        if type == SongType.STANDARD:
            return next((diff for diff in self.difficulties.standard if diff.level_index == level_index), None)
        if type == SongType.UTAGE:
            return next(iter(self.difficulties.utage), None)

    def get_difficulties(self, song_type: Union[SongType, _UnsetSentinel] = UNSET) -> Sequence["SongDifficulty"]:
        """Get all difficulties of the song, optionally filtered by type.

        Args:
            song_type: The type of the song (DX, STANDARD, UTAGE). If UNSET, returns all types.
        Returns:
            A sequence of difficulties matching the specified type.
        """
        if isinstance(song_type, _UnsetSentinel):
            return self.difficulties.standard + self.difficulties.dx + self.difficulties.utage
        if song_type == SongType.DX:
            return self.difficulties.dx
        if song_type == SongType.STANDARD:
            return self.difficulties.standard
        if song_type == SongType.UTAGE:
            return self.difficulties.utage

    def get_divingfish_id(self, type: SongType, level_index: LevelIndex) -> int:
        """Get the Diving Fish ID for a specific difficulty of this song.

        Args:
            type: The type of the song (DX, STANDARD, UTAGE).
            level_index: The level index of the difficulty.
        Returns:
            The Diving Fish ID for the specified difficulty.
        """
        difficulty = self.get_difficulty(type, level_index)
        if difficulty is None:
            raise ValueError(f"No difficulty found for type {type} and level index {level_index}")
        if difficulty.type == SongType.DX:
            return self.id + 10000
        if difficulty.type == SongType.UTAGE:
            return self.id + 100000
        return self.id

    def get_divingfish_ids(self, song_type: Union[SongType, _UnsetSentinel] = UNSET) -> set[int]:
        """Get a set of Diving Fish IDs for all difficulties of this song.

        Args:
            song_type: The type of the song (DX, STANDARD, UTAGE). If UNSET, returns IDs for all types.
        Returns:
            A set of Diving Fish IDs for the specified type or all types if UNSET.
        """
        ids = set()
        for difficulty in self.get_difficulties(song_type):
            ids.add(self.get_divingfish_id(difficulty.type, difficulty.level_index))
        return ids


@dataclass
class SongDifficulties:
    __slots__ = ("standard", "dx", "utage")

    standard: list["SongDifficulty"]
    dx: list["SongDifficulty"]
    utage: list["SongDifficultyUtage"]


@dataclass
class CurveObject:
    __slots__ = ("sample_size", "fit_level_value", "avg_achievements", "stdev_achievements", "avg_dx_score", "rate_sample_size", "fc_sample_size")

    sample_size: int
    fit_level_value: float
    avg_achievements: float
    stdev_achievements: float
    avg_dx_score: float
    rate_sample_size: dict[RateType, int]
    fc_sample_size: dict[FCType, int]


@dataclass
class SongDifficulty:
    __slots__ = (
        "id",
        "type",
        "level",
        "level_value",
        "level_index",
        "note_designer",
        "version",
        "tap_num",
        "hold_num",
        "slide_num",
        "touch_num",
        "break_num",
        "curve",
    )

    type: SongType
    level: str
    level_value: float
    level_index: LevelIndex
    note_designer: str
    version: int
    tap_num: int
    hold_num: int
    slide_num: int
    touch_num: int
    break_num: int
    curve: Optional[CurveObject]


@dataclass
class SongDifficultyUtage(SongDifficulty):
    __slots__ = ("kanji", "description", "is_buddy")

    kanji: str
    description: str
    is_buddy: bool


@dataclass
class PlayerIdentifier:
    qq: Optional[int] = None
    username: Optional[str] = None
    friend_code: Optional[int] = None
    credentials: Union[str, MutableMapping[str, Any], None] = None

    def __post_init__(self):
        if self.qq is None and self.username is None and self.friend_code is None and self.credentials is None:
            raise InvalidPlayerIdentifierError("At least one of the following must be provided: qq, username, friend_code, credentials")

    def _as_diving_fish(self) -> dict[str, Any]:
        if self.qq:
            return {"qq": str(self.qq)}
        elif self.username:
            return {"username": self.username}
        elif self.friend_code:
            raise InvalidPlayerIdentifierError("Friend code is not applicable for Diving Fish")
        else:
            raise InvalidPlayerIdentifierError("No valid identifier provided")

    def _as_lxns(self) -> str:
        if self.friend_code:
            return str(self.friend_code)
        elif self.qq:
            return f"qq/{str(self.qq)}"
        elif self.username:
            raise InvalidPlayerIdentifierError("Username is not applicable for LXNS")
        else:
            raise InvalidPlayerIdentifierError("No valid identifier provided")


@dataclass
class PlayerItem:
    @staticmethod
    def _namespace() -> str:
        raise NotImplementedError


@dataclass
class PlayerTrophy(PlayerItem):
    __slots__ = ("id", "name", "color")

    id: int
    name: str
    color: str

    @staticmethod
    def _namespace():
        return "trophies"


@dataclass
class PlayerIcon(PlayerItem):
    __slots__ = ("id", "name", "description", "genre")

    id: int
    name: str
    description: Optional[str]
    genre: Optional[str]

    @staticmethod
    def _namespace():
        return "icons"


@dataclass
class PlayerNamePlate(PlayerItem):
    __slots__ = ("id", "name", "description", "genre")

    id: int
    name: str
    description: Optional[str]
    genre: Optional[str]

    @staticmethod
    def _namespace():
        return "nameplates"


@dataclass
class PlayerFrame(PlayerItem):
    __slots__ = ("id", "name", "description", "genre")

    id: int
    name: str
    description: Optional[str]
    genre: Optional[str]

    @staticmethod
    def _namespace():
        return "frames"


@dataclass
class PlayerPartner(PlayerItem):
    __slots__ = ("id", "name")

    id: int
    name: str

    @staticmethod
    def _namespace():
        return "partners"


@dataclass
class PlayerChara(PlayerItem):
    __slots__ = ("id", "name")

    id: int
    name: str

    @staticmethod
    def _namespace():
        return "charas"


@dataclass
class PlayerRegion:
    __slots__ = ("region_id", "region_name", "play_count", "created_at")

    region_id: int
    region_name: str
    play_count: int
    created_at: datetime


@dataclass
class Player:
    __slots__ = ("identifier", "name", "rating")

    name: str
    rating: int


@dataclass
class DivingFishPlayer(Player):
    __slots__ = ("nickname", "plate", "additional_rating")

    nickname: str
    plate: str
    additional_rating: int


@dataclass
class LXNSPlayer(Player):
    __slots__ = ("friend_code", "course_rank", "class_rank", "star", "frame", "icon", "trophy", "name_plate", "upload_time")

    friend_code: int
    course_rank: int
    class_rank: int
    star: int
    frame: Optional[PlayerFrame]
    icon: Optional[PlayerIcon]
    trophy: Optional[PlayerTrophy]
    name_plate: Optional[PlayerNamePlate]
    upload_time: str


@dataclass
class ArcadePlayer(Player):
    __slots__ = ("is_login", "icon", "trophy", "name_plate")

    is_login: bool
    icon: Optional[PlayerIcon]
    trophy: Optional[PlayerTrophy]
    name_plate: Optional[PlayerNamePlate]


@dataclass
class AreaCharacter:
    __slots__ = ("id", "name", "illustrator", "description1", "description2", "team", "props")

    name: str
    illustrator: str
    description1: str
    description2: str
    team: str
    props: dict[str, str]


@dataclass
class AreaSong:
    __slots__ = ("id", "title", "artist", "description", "illustrator", "movie")

    id: Optional[int]
    title: str
    artist: str
    description: str
    illustrator: Optional[str]
    movie: Optional[str]


@dataclass
class Area:
    __slots__ = ("id", "name", "comment", "description", "video_id", "characters", "songs")

    id: str
    name: str
    comment: str
    description: str
    video_id: str
    characters: list[AreaCharacter]
    songs: list[AreaSong]


@dataclass
class Score:
    __slots__ = ("id", "level", "level_index", "achievements", "fc", "fs", "dx_score", "dx_rating", "play_count", "rate", "type", "song")

    id: int
    level: str
    level_index: LevelIndex
    achievements: Optional[float]
    fc: Optional[FCType]
    fs: Optional[FSType]
    dx_score: Optional[int]
    dx_rating: Optional[float]
    play_count: Optional[int]
    rate: RateType
    type: SongType

    def _compare(self, other: Optional["Score"]) -> "Score":
        if other is None:
            return self
        if self.dx_score != other.dx_score:  # larger value is better
            return self if (self.dx_score or 0) > (other.dx_score or 0) else other
        if self.achievements != other.achievements:  # larger value is better
            return self if (self.achievements or 0) > (other.achievements or 0) else other
        if self.rate != other.rate:  # smaller value is better
            self_rate = self.rate.value if self.rate is not None else 100
            other_rate = other.rate.value if other.rate is not None else 100
            return self if self_rate < other_rate else other
        if self.fc != other.fc:  # smaller value is better
            self_fc = self.fc.value if self.fc is not None else 100
            other_fc = other.fc.value if other.fc is not None else 100
            return self if self_fc < other_fc else other
        if self.fs != other.fs:  # bigger value is better
            self_fs = self.fs.value if self.fs is not None else -1
            other_fs = other.fs.value if other.fs is not None else -1
            return self if self_fs > other_fs else other
        return self  # we consider they are equal

    def _join(self, other: Optional["Score"]) -> "Score":
        if other is not None:
            if self.level_index != other.level_index or self.type != other.type:
                raise ValueError("Cannot join scores with different level indexes or types")
            self.achievements = max(self.achievements or 0, other.achievements or 0)
            if self.fc != other.fc:
                self_fc = self.fc.value if self.fc is not None else 100
                other_fc = other.fc.value if other.fc is not None else 100
                selected_value = min(self_fc, other_fc)
                self.fc = FCType(selected_value) if selected_value != 100 else None
            if self.fs != other.fs:
                self_fs = self.fs.value if self.fs is not None else -1
                other_fs = other.fs.value if other.fs is not None else -1
                selected_value = max(self_fs, other_fs)
                self.fs = FSType(selected_value) if selected_value != -1 else None
            if self.rate != other.rate:
                selected_value = min(self.rate.value, other.rate.value)
                self.rate = RateType(selected_value)
        return self


@dataclass
class ScoreExtend(Score):
    __slots__ = ["title", "level_value", "level_dx_score"]

    title: str
    level_value: float
    level_dx_score: int


@dataclass
class PlateObject:
    __slots__ = ("song", "levels", "scores")

    song: Song
    levels: set[LevelIndex]
    scores: list[ScoreExtend]


@dataclass
class PlayerSong:
    __slots__ = ["song", "scores"]

    song: Song
    scores: list[ScoreExtend]


@dataclass
class PlayerBests:
    __slots__ = ["rating", "rating_b35", "rating_b15", "scores_b35", "scores_b15"]

    rating: int
    rating_b35: int
    rating_b15: int
    scores_b35: list[ScoreExtend]
    scores_b15: list[ScoreExtend]

    @property
    def scores(self) -> list[ScoreExtend]:
        """Get all scores, including both B35 and B15."""
        return self.scores_b35 + self.scores_b15



from typing import Any, TypeVar

_T = TypeVar("_T")


class _UnsetSentinel:
    def __repr__(self) -> str:
        return "Unset"

    def __copy__(self: _T) -> _T:
        return self

    def __reduce__(self) -> str:
        return "Unset"

    def __deepcopy__(self: _T, _: Any) -> _T:
        return self


UNSET = _UnsetSentinel()



import os
from importlib.util import find_spec
from typing import Annotated, Any, Callable, Literal, Optional, Union
from urllib.parse import unquote, urlparse

from maimai_py.maimai import MaimaiClient, MaimaiClientMultithreading, MaimaiPlates, MaimaiSongs, _UnsetSentinel
from maimai_py.models import *
from maimai_py.providers import *
from maimai_py.providers.hybrid import HybridProvider

PlateAttrs = Literal["remained", "cleared", "played", "all"]


def xstr(s: Optional[str]) -> str:
    return "" if s is None else str(s).lower()


def istr(i: Optional[list]) -> str:
    return "" if i is None else "".join(i).lower()


def pagination(page_size, page, data):
    total_pages = (len(data) + page_size - 1) // page_size
    if page < 1 or page > total_pages:
        return []

    start = (page - 1) * page_size
    end = page * page_size
    return data[start:end]


def get_filters(functions: dict[Any, Callable[..., bool]]):
    union = [flag for cond, flag in functions.items() if cond is not None]
    filter = lambda obj: all([flag(obj) for flag in union])
    return filter


if find_spec("fastapi"):
    from fastapi import APIRouter, Depends, FastAPI, Query, Request
    from fastapi.openapi.utils import get_openapi
    from fastapi.responses import JSONResponse

    class MaimaiRoutes:
        _client: MaimaiClient

        _lxns_token: Optional[str] = None
        _divingfish_token: Optional[str] = None
        _arcade_proxy: Optional[str] = None

        def __init__(
            self,
            client: MaimaiClient,
            lxns_token: Optional[str] = None,
            divingfish_token: Optional[str] = None,
            arcade_proxy: Optional[str] = None,
        ):
            self._client = client
            self._lxns_token = lxns_token
            self._divingfish_token = divingfish_token
            self._arcade_proxy = arcade_proxy

        def _dep_lxns_player(self, credentials: Optional[str] = None, friend_code: Optional[int] = None, qq: Optional[int] = None):
            return PlayerIdentifier(credentials=credentials, qq=qq, friend_code=friend_code)

        def _dep_divingfish_player(self, username: Optional[str] = None, credentials: Optional[str] = None, qq: Optional[int] = None):
            return PlayerIdentifier(qq=qq, credentials=credentials, username=username)

        def _dep_arcade_player(self, credentials: str):
            return PlayerIdentifier(credentials=credentials)

        def _dep_divingfish(self) -> IProvider:
            return DivingFishProvider(developer_token=self._divingfish_token)

        def _dep_lxns(self) -> IProvider:
            return LXNSProvider(developer_token=self._lxns_token)

        def _dep_arcade(self) -> IProvider:
            return ArcadeProvider(http_proxy=self._arcade_proxy)

        def _dep_hybrid(self) -> IProvider:
            return HybridProvider()

        def get_router(self, dep_provider: Callable, dep_player: Optional[Callable] = None, skip_base: bool = True) -> APIRouter:
            router = APIRouter()

            def try_add_route(func: Callable, router: APIRouter, dep_provider: Callable):
                provider_type = func.__annotations__.get("provider")
                if provider_type and isinstance(dep_provider(), provider_type):
                    method = "GET" if "get_" in func.__name__ else "POST"
                    response_model = func.__annotations__.get("return")
                    router.add_api_route(
                        f"/{func.__name__.split('_')[-1]}",
                        func,
                        name=f"{func.__name__}",
                        methods=[method],
                        response_model=response_model,
                        description=func.__doc__,
                    )

            async def _get_songs(
                id: Optional[int] = None,
                title: Optional[str] = None,
                artist: Optional[str] = None,
                genre: Optional[Genre] = None,
                bpm: Optional[int] = None,
                map: Optional[str] = None,
                version: Optional[int] = None,
                type: Optional[SongType] = None,
                level: Optional[str] = None,
                versions: Optional[Version] = None,
                keywords: Optional[str] = None,
                page: int = Query(1, ge=1),
                page_size: int = Query(100, ge=1),
                provider: ISongProvider = Depends(dep_provider),
            ) -> list[Song]:
                maimai_songs: MaimaiSongs = await self._client.songs(provider=provider)
                type_func: Callable[[Song], bool] = lambda song: song.get_difficulties(type) != []  # type: ignore
                level_func: Callable[[Song], bool] = lambda song: any([diff.level == level for diff in song.get_difficulties()])
                versions_func: Callable[[Song], bool] = lambda song: versions.value <= song.version < all_versions[all_versions.index(versions) + 1].value  # type: ignore
                keywords_func: Callable[[Song], bool] = lambda song: xstr(keywords) in xstr(song.title) + xstr(song.artist) + istr(song.aliases)
                songs = await maimai_songs.filter(id=id, title=title, artist=artist, genre=genre, bpm=bpm, map=map, version=version)
                filters = get_filters({type: type_func, level: level_func, versions: versions_func, keywords: keywords_func})
                result = [song for song in songs if filters(song)]
                return pagination(page_size, page, result)

            async def _get_icons(
                id: Optional[int] = None,
                name: Optional[str] = None,
                description: Optional[str] = None,
                genre: Optional[str] = None,
                keywords: Optional[str] = None,
                page: int = Query(1, ge=1),
                page_size: int = Query(100, ge=1),
                provider: IItemListProvider = Depends(dep_provider),
            ) -> list[PlayerIcon]:
                items = await self._client.items(PlayerIcon, provider=provider)
                if id is not None:
                    return [item] if (item := await items.by_id(id)) else []
                keyword_func = lambda icon: xstr(keywords) in (xstr(icon.name) + xstr(icon.description) + xstr(icon.genre))
                filters = get_filters({keywords: keyword_func})
                result = [x for x in await items.filter(name=name, description=description, genre=genre) if filters(x)]
                return pagination(page_size, page, result)

            async def _get_nameplates(
                id: Optional[int] = None,
                name: Optional[str] = None,
                description: Optional[str] = None,
                genre: Optional[str] = None,
                keywords: Optional[str] = None,
                page: int = Query(1, ge=1),
                page_size: int = Query(100, ge=1),
                provider: IItemListProvider = Depends(dep_provider),
            ) -> list[PlayerNamePlate]:
                items = await self._client.items(PlayerNamePlate, provider=provider)
                if id is not None:
                    return [item] if (item := await items.by_id(id)) else []
                keyword_func = lambda icon: xstr(keywords) in (xstr(icon.name) + xstr(icon.description) + xstr(icon.genre))
                filters = get_filters({keywords: keyword_func})
                result = [x for x in await items.filter(name=name, description=description, genre=genre) if filters(x)]
                return pagination(page_size, page, result)

            async def _get_frames(
                id: Optional[int] = None,
                name: Optional[str] = None,
                description: Optional[str] = None,
                genre: Optional[str] = None,
                keywords: Optional[str] = None,
                page: int = Query(1, ge=1),
                page_size: int = Query(100, ge=1),
                provider: IItemListProvider = Depends(dep_provider),
            ) -> list[PlayerFrame]:
                items = await self._client.items(PlayerFrame, provider=provider)
                if id is not None:
                    return [item] if (item := await items.by_id(id)) else []
                keyword_func = lambda icon: xstr(keywords) in (xstr(icon.name) + xstr(icon.description) + xstr(icon.genre))
                filters = get_filters({keywords: keyword_func})
                result = [x for x in await items.filter(name=name, description=description, genre=genre) if filters(x)]
                return pagination(page_size, page, result)

            async def _get_trophies(
                id: Optional[int] = None,
                name: Optional[str] = None,
                color: Optional[str] = None,
                keywords: Optional[str] = None,
                page: int = Query(1, ge=1),
                page_size: int = Query(100, ge=1),
                provider: IItemListProvider = Depends(dep_provider),
            ) -> list[PlayerTrophy]:
                items = await self._client.items(PlayerTrophy, provider=provider)
                if id is not None:
                    return [item] if (item := await items.by_id(id)) else []
                keyword_func = lambda icon: xstr(keywords) in (xstr(icon.name) + xstr(icon.color))
                filters = get_filters({keywords: keyword_func})
                result = [x for x in await items.filter(name=name, color=color) if filters(x)]
                return pagination(page_size, page, result)

            async def _get_charas(
                id: Optional[int] = None,
                name: Optional[str] = None,
                keywords: Optional[str] = None,
                page: int = Query(1, ge=1),
                page_size: int = Query(100, ge=1),
                provider: IItemListProvider = Depends(dep_provider),
            ) -> list[PlayerChara]:
                items = await self._client.items(PlayerChara, provider=provider)
                if id is not None:
                    return [item] if (item := await items.by_id(id)) else []
                keyword_func = lambda chara: xstr(keywords) in xstr(chara.name)
                filters = get_filters({keywords: keyword_func})
                result = [x for x in await items.filter(name=name) if filters(x)]
                return pagination(page_size, page, result)

            async def _get_partners(
                id: Optional[int] = None,
                name: Optional[str] = None,
                keywords: Optional[str] = None,
                page: int = Query(1, ge=1),
                page_size: int = Query(100, ge=1),
                provider: IItemListProvider = Depends(dep_provider),
            ) -> list[PlayerPartner]:
                items = await self._client.items(PlayerPartner, provider=provider)
                if id is not None:
                    return [item] if (item := await items.by_id(id)) else []
                keyword_func = lambda partner: xstr(keywords) in xstr(partner.name)
                filters = get_filters({keywords: keyword_func})
                result = [x for x in await items.filter(name=name) if filters(x)]
                return pagination(page_size, page, result)

            async def _get_areas(
                lang: Literal["ja", "zh"] = "ja",
                id: Optional[str] = None,
                name: Optional[str] = None,
                keywords: Optional[str] = None,
                page: int = Query(1, ge=1),
                page_size: int = Query(100, ge=1),
                provider: IAreaProvider = Depends(dep_provider),
            ) -> list[Area]:
                areas = await self._client.areas(lang, provider=provider)
                if id is not None:
                    return [area] if (area := await areas.by_id(id)) else []
                if name is not None:
                    return [area] if (area := await areas.by_name(name)) else []
                keyword_func = lambda area: xstr(keywords) in (xstr(area.name) + xstr(area.comment))
                filters = get_filters({keywords: keyword_func})
                result = [x for x in await areas.get_all() if filters(x)]
                return pagination(page_size, page, result)

            async def _get_scores(
                provider: IScoreProvider = Depends(dep_provider),
                player: PlayerIdentifier = Depends(dep_player),
            ) -> list[ScoreExtend]:
                scores = await self._client.scores(player, provider=provider)
                return scores.scores

            async def _get_regions(
                provider: IRegionProvider = Depends(dep_provider),
                player: PlayerIdentifier = Depends(dep_player),
            ) -> list[PlayerRegion]:
                return await self._client.regions(player, provider=provider)

            async def _get_players(
                provider: IPlayerProvider = Depends(dep_provider),
                player: PlayerIdentifier = Depends(dep_player),
            ) -> Union[Player, DivingFishPlayer, LXNSPlayer, ArcadePlayer]:
                return await self._client.players(player, provider=provider)

            async def _get_bests(
                provider: IScoreProvider = Depends(dep_provider),
                player: PlayerIdentifier = Depends(dep_player),
            ) -> PlayerBests:
                maimai_scores = await self._client.bests(player, provider=provider)
                return maimai_scores.get_player_bests()

            async def _post_scores(
                scores: list[Score],
                provider: IScoreUpdateProvider = Depends(dep_provider),
                player: PlayerIdentifier = Depends(dep_player),
            ) -> None:
                await self._client.updates(player, scores, provider=provider)

            async def _get_plates(
                plate: str,
                attr: Literal["remained", "cleared", "played", "all"] = "remained",
                provider: IScoreProvider = Depends(dep_provider),
                player: PlayerIdentifier = Depends(dep_player),
            ) -> list[PlateObject]:
                plates: MaimaiPlates = await self._client.plates(player, plate, provider=provider)
                return await getattr(plates, f"get_{attr}")()

            async def _get_minfo(
                id: Optional[int] = None,
                title: Optional[str] = None,
                keywords: Optional[str] = None,
                provider: IScoreProvider = Depends(dep_provider),
                player: PlayerIdentifier = Depends(dep_player),
            ) -> Optional[PlayerSong]:
                song_trait = id if id is not None else title if title is not None else keywords if keywords is not None else None
                if song_trait is not None:
                    return await self._client.minfo(song_trait, player, provider=provider)

            async def _get_identifiers(
                code: str,
                provider: IPlayerIdentifierProvider = Depends(dep_provider),
            ) -> PlayerIdentifier:
                return await self._client.identifiers(code, provider=provider)

            bases: list[Callable] = [_get_songs, _get_icons, _get_nameplates, _get_frames, _get_trophies, _get_charas, _get_partners, _get_areas]
            players: list[Callable] = [_get_scores, _get_regions, _get_players, _get_bests, _post_scores, _get_plates, _get_minfo, _get_identifiers]

            all = players + (bases if not skip_base else [])
            [try_add_route(func, router, dep_provider) for func in all]

            return router


if all([find_spec(p) for p in ["fastapi", "uvicorn", "typer"]]):
    import typer
    import uvicorn
    from fastapi import APIRouter, Depends, FastAPI, Request
    from fastapi.openapi.utils import get_openapi
    from fastapi.responses import JSONResponse

    # prepare for ASGI app
    asgi_app = FastAPI(title="maimai.py API", description="The definitive python wrapper for MaimaiCN related development.")
    routes = MaimaiRoutes(MaimaiClientMultithreading())  # type: ignore

    # register routes and middlewares
    asgi_app.include_router(routes.get_router(routes._dep_hybrid, skip_base=False), tags=["base"])
    asgi_app.include_router(routes.get_router(routes._dep_divingfish, routes._dep_divingfish_player), prefix="/divingfish", tags=["divingfish"])
    asgi_app.include_router(routes.get_router(routes._dep_lxns, routes._dep_lxns_player), prefix="/lxns", tags=["lxns"])
    asgi_app.include_router(routes.get_router(routes._dep_arcade, routes._dep_arcade_player), prefix="/arcade", tags=["arcade"])

    def main(
        host: Annotated[str, typer.Option(help="The host address to bind to.")] = "127.0.0.1",
        port: Annotated[int, typer.Option(help="The port number to bind to.")] = 8000,
        redis: Annotated[Optional[str], typer.Option(help="Redis server address, for example: redis://localhost:6379/0.")] = None,
        lxns_token: Annotated[Optional[str], typer.Option(help="LXNS token for LXNS API.")] = None,
        divingfish_token: Annotated[Optional[str], typer.Option(help="Diving Fish token for Diving Fish API.")] = None,
        arcade_proxy: Annotated[Optional[str], typer.Option(help="HTTP proxy for Arcade API.")] = None,
    ):
        # prepare for redis cache backend
        redis_backend = UNSET
        if redis and find_spec("redis"):
            from aiocache import RedisCache
            from aiocache.serializers import PickleSerializer

            redis_url = urlparse(redis)
            redis_backend = RedisCache(
                serializer=PickleSerializer(),
                endpoint=unquote(redis_url.hostname or "localhost"),
                port=redis_url.port or 6379,
                password=redis_url.password,
                db=int(unquote(redis_url.path).replace("/", "")),
            )

        # override the default maimai.py client
        routes._client._cache = routes._client._cache if isinstance(redis_backend, _UnsetSentinel) else redis_backend
        routes._lxns_token = lxns_token or os.environ.get("LXNS_DEVELOPER_TOKEN")
        routes._divingfish_token = divingfish_token or os.environ.get("DIVINGFISH_DEVELOPER_TOKEN")
        routes._arcade_proxy = arcade_proxy

        @asgi_app.exception_handler(MaimaiPyError)
        async def exception_handler(request: Request, exc: MaimaiPyError):
            return JSONResponse(
                status_code=400,
                content={"message": f"Oops! There goes a maimai.py error {exc}.", "details": repr(exc)},
            )

        @asgi_app.get("/", include_in_schema=False)
        async def root():
            return {"message": "Hello, maimai.py! Check /docs for more information."}

        # run the ASGI app with uvicorn
        uvicorn.run(asgi_app, host=host, port=port)

    def openapi():
        specs = get_openapi(
            title=asgi_app.title,
            version=asgi_app.version,
            openapi_version=asgi_app.openapi_version,
            description=asgi_app.description,
            routes=asgi_app.routes,
        )
        with open(f"openapi.json", "w") as f:
            json.dump(specs, f)

    if __name__ == "__main__":
        typer.run(main)


if find_spec("maimai_ffi") and find_spec("nuitka"):
    import json

    import cryptography
    import cryptography.fernet
    import cryptography.hazmat.backends
    import cryptography.hazmat.primitives.ciphers
    import maimai_ffi
    import maimai_ffi.model
    import maimai_ffi.request
    import redis



from importlib.util import find_spec

from .maimai import MaimaiAreas, MaimaiClient, MaimaiClientMultithreading, MaimaiItems, MaimaiPlates, MaimaiScores, MaimaiSongs
from .models import *
from .providers import *

if find_spec("fastapi"):
    from .api import MaimaiRoutes



import asyncio
import dataclasses
import hashlib
from functools import reduce
from json import JSONDecodeError
from operator import concat
from typing import TYPE_CHECKING, Iterable

from httpx import HTTPStatusError, Response

from maimai_py.models import *
from maimai_py.models import PlayerIdentifier, Score, Song

from .base import IAliasProvider, IItemListProvider, IPlayerProvider, IScoreProvider, IScoreUpdateProvider, ISongProvider

if TYPE_CHECKING:
    from maimai_py.maimai import MaimaiClient, MaimaiSongs


class LXNSProvider(ISongProvider, IPlayerProvider, IScoreProvider, IScoreUpdateProvider, IAliasProvider, IItemListProvider):
    """The provider that fetches data from the LXNS.

    LXNS: https://maimai.lxns.net/
    """

    developer_token: Optional[str]
    """The developer token used to access the LXNS API."""
    base_url = "https://maimai.lxns.net/"
    """The base URL for the LXNS API."""

    @property
    def headers(self):
        """@private"""
        if not self.developer_token:
            raise InvalidDeveloperTokenError("Developer token is not provided.")
        return {"Authorization": self.developer_token}

    def __init__(self, developer_token: Optional[str] = None):
        """Initializes the LXNSProvider.

        Args:
            developer_token: The developer token used to access the LXNS API.
        """
        self.developer_token = developer_token

    def _hash(self) -> str:
        return hashlib.md5(b"lxns").hexdigest()

    async def _ensure_friend_code(self, client: "MaimaiClient", identifier: PlayerIdentifier) -> None:
        if identifier.friend_code is None:
            if identifier.qq is not None:
                resp = await client._client.get(self.base_url + f"api/v0/maimai/player/qq/{identifier.qq}", headers=self.headers)
                if not resp.json()["success"]:
                    raise InvalidPlayerIdentifierError(resp.json()["message"])
                identifier.friend_code = resp.json()["data"]["friend_code"]

    async def _build_player_request(self, path: str, identifier: PlayerIdentifier, client: "MaimaiClient") -> tuple[str, dict[str, str], bool]:
        use_user_api = identifier.credentials is not None and isinstance(identifier.credentials, str)
        if use_user_api:
            # user-level API takes the precedence: If personal token provided, use it first
            assert isinstance(identifier.credentials, str)
            entrypoint = f"api/v0/user/maimai/player/{path}"
            headers = {"X-User-Token": identifier.credentials}
        else:
            await self._ensure_friend_code(client, identifier)
            entrypoint = f"api/v0/maimai/player/{identifier.friend_code}/{path}"
            headers = self.headers
        entrypoint = entrypoint.removesuffix("/")
        return self.base_url + entrypoint, headers, use_user_api

    @staticmethod
    def _deser_note(diff: dict, key: str) -> int:
        if "notes" in diff:
            if "is_buddy" in diff and diff["is_buddy"]:
                return diff["notes"]["left"][key] + diff["notes"]["right"][key]
            return diff["notes"][key]
        return 0

    @staticmethod
    def _deser_item(item: dict, cls: type) -> Any:
        return cls(
            id=item["id"],
            name=item["name"],
            description=item["description"] if "description" in item else None,
            genre=item["genre"] if "genre" in item else None,
        )

    @staticmethod
    def _deser_song(song: dict) -> Song:
        return Song(
            id=song["id"],
            title=song["title"],
            artist=song["artist"],
            genre=name_to_genre[song["genre"]],
            bpm=song["bpm"],
            aliases=song["aliases"] if "aliases" in song else None,
            map=song["map"] if "map" in song else None,
            version=song["version"],
            rights=song["rights"] if "rights" in song else None,
            disabled=song["disabled"] if "disabled" in song else False,
            difficulties=SongDifficulties(standard=[], dx=[], utage=[]),
        )

    @staticmethod
    def _deser_diff(difficulty: dict) -> SongDifficulty:
        return SongDifficulty(
            type=SongType[difficulty["type"].upper()],
            level=difficulty["level"],
            level_value=difficulty["level_value"],
            level_index=LevelIndex(difficulty["difficulty"]),
            note_designer=difficulty["note_designer"],
            version=difficulty["version"],
            tap_num=LXNSProvider._deser_note(difficulty, "tap"),
            hold_num=LXNSProvider._deser_note(difficulty, "hold"),
            slide_num=LXNSProvider._deser_note(difficulty, "slide"),
            touch_num=LXNSProvider._deser_note(difficulty, "touch"),
            break_num=LXNSProvider._deser_note(difficulty, "break"),
            curve=None,
        )

    @staticmethod
    def _deser_diff_utage(difficulty: dict) -> SongDifficultyUtage:
        return SongDifficultyUtage(
            **dataclasses.asdict(LXNSProvider._deser_diff(difficulty)),
            kanji=difficulty["kanji"],
            description=difficulty["description"],
            is_buddy=difficulty["is_buddy"],
        )

    @staticmethod
    def _deser_score(score: dict) -> Score:
        return Score(
            id=score["id"],
            level=score["level"],
            level_index=LevelIndex(score["level_index"]),
            achievements=score["achievements"] if "achievements" in score else None,
            fc=FCType[score["fc"].upper()] if score["fc"] else None,
            fs=FSType[score["fs"].upper()] if score["fs"] else None,
            dx_score=score["dx_score"] if "dx_score" in score else None,
            dx_rating=int(score["dx_rating"]) if "dx_rating" in score else None,
            play_count=None,
            rate=RateType[score["rate"].upper()],
            type=SongType[score["type"].upper()],
        )

    @staticmethod
    async def _ser_score(score: Score, songs: "MaimaiSongs") -> Optional[dict]:
        song_title = song.title if (song := await songs.by_id(score.id)) else None
        if song_title is not None:
            return {
                "id": score.id,
                "song_name": song_title,
                "level": score.level,
                "level_index": score.level_index.value,
                "achievements": score.achievements,
                "fc": score.fc.name.lower() if score.fc else None,
                "fs": score.fs.name.lower() if score.fs else None,
                "dx_score": score.dx_score,
                "dx_rating": score.dx_rating,
                "rate": score.rate.name.lower(),
                "type": score.type.name.lower(),
            }

    def _check_response_player(self, resp: Response) -> dict:
        try:
            resp_json = resp.json()
            if not resp_json["success"]:
                if resp_json["code"] in [400, 404]:
                    raise InvalidPlayerIdentifierError(resp_json["message"])
                elif resp_json["code"] in [403]:
                    raise PrivacyLimitationError(resp_json["message"])
                elif resp_json["code"] in [401]:
                    raise InvalidDeveloperTokenError(resp_json["message"])
                elif resp.status_code in [400, 401]:
                    raise InvalidPlayerIdentifierError(resp_json["message"])
                elif not resp.is_success:
                    resp.raise_for_status()
            return resp_json
        except JSONDecodeError as exc:
            raise InvalidJsonError(resp.text) from exc
        except HTTPStatusError as exc:
            raise MaimaiPyError(exc) from exc

    async def get_songs(self, client: "MaimaiClient") -> list[Song]:
        resp = await client._client.get(self.base_url + "api/v0/maimai/song/list?notes=true")
        resp.raise_for_status()
        resp_json = resp.json()
        unique_songs: dict[int, Song] = {}
        for song in resp_json["songs"]:
            unique_key = int(song["id"]) % 10000
            if unique_key not in unique_songs:
                unique_songs[unique_key] = LXNSProvider._deser_song(song)
            difficulties = unique_songs[unique_key].difficulties
            difficulties.standard.extend(LXNSProvider._deser_diff(difficulty) for difficulty in song["difficulties"].get("standard", []))
            difficulties.dx.extend(LXNSProvider._deser_diff(difficulty) for difficulty in song["difficulties"].get("dx", []))
            difficulties.utage.extend(LXNSProvider._deser_diff_utage(difficulty) for difficulty in song["difficulties"].get("utage", []))
        return list(unique_songs.values())

    async def get_player(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> LXNSPlayer:
        maimai_frames = await client.items(PlayerFrame)
        maimai_icons = await client.items(PlayerIcon)
        maimai_trophies = await client.items(PlayerTrophy)
        maimai_nameplates = await client.items(PlayerNamePlate)
        url, headers, _ = await self._build_player_request("", identifier, client)
        resp = await client._client.get(url, headers=headers)
        resp_data = self._check_response_player(resp)["data"]
        return LXNSPlayer(
            name=resp_data["name"],
            rating=resp_data["rating"],
            friend_code=resp_data["friend_code"],
            course_rank=resp_data["course_rank"],
            class_rank=resp_data["class_rank"],
            star=resp_data["star"],
            frame=await maimai_frames.by_id(resp_data["frame"]["id"]) if "frame" in resp_data else None,
            icon=await maimai_icons.by_id(resp_data["icon"]["id"]) if "icon" in resp_data else None,
            trophy=await maimai_trophies.by_id(resp_data["trophy"]["id"]) if "trophy" in resp_data else None,
            name_plate=await maimai_nameplates.by_id(resp_data["name_plate"]["id"]) if "name_plate" in resp_data else None,
            upload_time=resp_data["upload_time"],
        )

    async def get_scores_all(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[Score]:
        url, headers, use_user_api = await self._build_player_request("scores", identifier, client)
        resp = await client._client.get(url, headers=headers)
        resp_data = self._check_response_player(resp)["data"]
        scores = [s for score in resp_data if (s := LXNSProvider._deser_score(score))]
        if not use_user_api:
            # LXNSProvider's developer-level API scores are incomplete, which doesn't contain dx_rating and achievements, leading to sorting difficulties.
            # In this case, we should always fetch the b35 and b15 scores for LXNSProvider.
            scores.extend(await self.get_scores_best(identifier, client))
        return scores

    async def get_scores_best(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[Score]:
        if identifier.friend_code is None:
            return await self.get_scores_all(identifier, client)
        await self._ensure_friend_code(client, identifier)
        entrypoint = f"api/v0/maimai/player/{identifier.friend_code}/bests"
        resp = await client._client.get(self.base_url + entrypoint, headers=self.headers)
        resp_data = self._check_response_player(resp)["data"]
        return [s for score in resp_data["standard"] + resp_data["dx"] if (s := LXNSProvider._deser_score(score))]

    async def get_scores_one(self, identifier: PlayerIdentifier, song: Song, client: "MaimaiClient") -> list[Score]:
        await self._ensure_friend_code(client, identifier)
        request_tasks, create_task = [], lambda type: asyncio.create_task(
            client._client.get(
                self.base_url + f"api/v0/maimai/player/{identifier.friend_code}/bests",
                params={"song_id": song.id if type != SongType.UTAGE else song.id + 100000, "song_type": type.value},
                headers=self.headers,
            )
        )
        if len(song.difficulties.standard) > 0:
            request_tasks.append(create_task(SongType.STANDARD))
        if len(song.difficulties.dx) > 0:
            request_tasks.append(create_task(SongType.DX))
        if len(song.difficulties.utage) > 0:
            request_tasks.append(create_task(SongType.UTAGE))
        resps = await asyncio.gather(*request_tasks)
        resp_data = [self._check_response_player(resp)["data"] for resp in resps]
        return [s for score in reduce(concat, resp_data) if (s := LXNSProvider._deser_score(score))]

    async def update_scores(self, identifier: PlayerIdentifier, scores: Iterable[Score], client: "MaimaiClient") -> None:
        maimai_songs = await client.songs()
        url, headers, _ = await self._build_player_request("scores", identifier, client)
        scores_dict = {"scores": [json for score in scores if (json := await LXNSProvider._ser_score(score, maimai_songs))]}
        resp = await client._client.post(url, headers=headers, json=scores_dict)
        resp.raise_for_status()
        resp_json = resp.json()
        if not resp_json["success"] and resp_json["code"] == 400:
            raise ValueError(resp_json["message"])

    async def get_aliases(self, client: "MaimaiClient") -> dict[int, list[str]]:
        resp = await client._client.get(self.base_url + "api/v0/maimai/alias/list")
        resp.raise_for_status()
        return {item["song_id"]: item["aliases"] for item in resp.json()["aliases"]}

    async def get_icons(self, client: "MaimaiClient") -> dict[int, PlayerIcon]:
        resp = await client._client.get(self.base_url + "api/v0/maimai/icon/list")
        resp.raise_for_status()
        return {item["id"]: LXNSProvider._deser_item(item, PlayerIcon) for item in resp.json()["icons"]}

    async def get_nameplates(self, client: "MaimaiClient") -> dict[int, PlayerNamePlate]:
        resp = await client._client.get(self.base_url + "api/v0/maimai/plate/list")
        resp.raise_for_status()
        return {item["id"]: LXNSProvider._deser_item(item, PlayerNamePlate) for item in resp.json()["plates"]}

    async def get_frames(self, client: "MaimaiClient") -> dict[int, PlayerFrame]:
        resp = await client._client.get(self.base_url + "api/v0/maimai/frame/list")
        resp.raise_for_status()
        return {item["id"]: LXNSProvider._deser_item(item, PlayerFrame) for item in resp.json()["frames"]}



from enum import Enum, IntEnum
from typing import Union

from maimai_ffi.model import region_map


class Version(IntEnum):
    MAIMAI = 10000
    MAIMAI_PLUS = 11000
    MAIMAI_GREEN = 12000
    MAIMAI_GREEN_PLUS = 13000
    MAIMAI_ORANGE = 14000
    MAIMAI_ORANGE_PLUS = 15000
    MAIMAI_PINK = 16000
    MAIMAI_PINK_PLUS = 17000
    MAIMAI_MURASAKI = 18000
    MAIMAI_MURASAKI_PLUS = 18500
    MAIMAI_MILK = 19000
    MAIMAI_MILK_PLUS = 19500
    MAIMAI_FINALE = 19900
    MAIMAI_DX = 20000  # 舞萌DX
    MAIMAI_DX_SPLASH = 21000  # 舞萌DX 2021
    MAIMAI_DX_UNIVERSE = 22000  # 舞萌DX 2022
    MAIMAI_DX_FESTIVAL = 23000  # 舞萌DX 2023
    MAIMAI_DX_BUDDIES = 24000  # 舞萌DX 2024
    MAIMAI_DX_PRISM = 25000  # 舞萌DX 2025
    MAIMAI_DX_FUTURE = 30000  # 舞萌DX 2077


class Genre(Enum):
    POPSアニメ = "POPSアニメ"
    niconicoボーカロイド = "niconicoボーカロイド"
    東方Project = "東方Project"
    ゲームバラエティ = "ゲームバラエティ"
    maimai = "maimai"
    オンゲキCHUNITHM = "オンゲキCHUNITHM"
    宴会場 = "宴会場"


region_map: tuple[str, ...] = region_map

all_versions = list(Version.__members__.values())
"""@private"""

current_version = all_versions[-2]
"""@private"""

plate_to_version: dict[str, Version] = {
    "初": Version.MAIMAI,
    "真": Version.MAIMAI_PLUS,
    "超": Version.MAIMAI_GREEN,
    "檄": Version.MAIMAI_GREEN_PLUS,
    "橙": Version.MAIMAI_ORANGE,
    "晓": Version.MAIMAI_ORANGE_PLUS,
    "桃": Version.MAIMAI_PINK,
    "樱": Version.MAIMAI_PINK_PLUS,
    "紫": Version.MAIMAI_MURASAKI,
    "堇": Version.MAIMAI_MURASAKI_PLUS,
    "白": Version.MAIMAI_MILK,
    "雪": Version.MAIMAI_MILK_PLUS,
    "辉": Version.MAIMAI_FINALE,
    "熊": Version.MAIMAI_DX,
    "华": Version.MAIMAI_DX,
    "爽": Version.MAIMAI_DX_SPLASH,
    "煌": Version.MAIMAI_DX_SPLASH,
    "星": Version.MAIMAI_DX_UNIVERSE,
    "宙": Version.MAIMAI_DX_UNIVERSE,
    "祭": Version.MAIMAI_DX_FESTIVAL,
    "祝": Version.MAIMAI_DX_FESTIVAL,
    "双": Version.MAIMAI_DX_BUDDIES,
    "宴": Version.MAIMAI_DX_BUDDIES,
    "镜": Version.MAIMAI_DX_PRISM,
    "未": Version.MAIMAI_DX_FUTURE,
}
"""@private"""


divingfish_to_version: dict[str, Version] = {
    "maimai": Version.MAIMAI,
    "maimai PLUS": Version.MAIMAI_PLUS,
    "maimai GreeN": Version.MAIMAI_GREEN,
    "maimai GreeN PLUS": Version.MAIMAI_GREEN_PLUS,
    "maimai ORANGE": Version.MAIMAI_ORANGE,
    "maimai ORANGE PLUS": Version.MAIMAI_ORANGE_PLUS,
    "maimai PiNK": Version.MAIMAI_PINK,
    "maimai PiNK PLUS": Version.MAIMAI_PINK_PLUS,
    "maimai MURASAKi": Version.MAIMAI_MURASAKI,
    "maimai MURASAKi PLUS": Version.MAIMAI_MURASAKI_PLUS,
    "maimai MiLK": Version.MAIMAI_MILK,
    "MiLK PLUS": Version.MAIMAI_MILK_PLUS,
    "maimai FiNALE": Version.MAIMAI_FINALE,
    "maimai でらっくす": Version.MAIMAI_DX,
    "maimai でらっくす PLUS": Version.MAIMAI_DX,
    "maimai でらっくす Splash": Version.MAIMAI_DX_SPLASH,
    "maimai でらっくす Splash PLUS": Version.MAIMAI_DX_SPLASH,
    "maimai でらっくす UNiVERSE": Version.MAIMAI_DX_UNIVERSE,
    "maimai でらっくす UNiVERSE PLUS": Version.MAIMAI_DX_UNIVERSE,
    "maimai でらっくす FESTiVAL": Version.MAIMAI_DX_FESTIVAL,
    "maimai でらっくす FESTiVAL PLUS": Version.MAIMAI_DX_FESTIVAL,
    "maimai でらっくす BUDDiES": Version.MAIMAI_DX_BUDDIES,
    "maimai でらっくす BUDDiES PLUS": Version.MAIMAI_DX_BUDDIES,
    "maimai でらっくす PRiSM": Version.MAIMAI_DX_PRISM,
}
"""@private"""

name_to_genre: dict[str, Genre] = {
    "POPSアニメ": Genre.POPSアニメ,
    "流行&动漫": Genre.POPSアニメ,
    "niconicoボーカロイド": Genre.niconicoボーカロイド,
    "niconico & VOCALOID": Genre.niconicoボーカロイド,
    "東方Project": Genre.東方Project,
    "东方Project": Genre.東方Project,
    "ゲームバラエティ": Genre.ゲームバラエティ,
    "其他游戏": Genre.ゲームバラエティ,
    "maimai": Genre.maimai,
    "舞萌": Genre.maimai,
    "オンゲキCHUNITHM": Genre.オンゲキCHUNITHM,
    "音击&中二节奏": Genre.オンゲキCHUNITHM,
    "宴会場": Genre.宴会場,
}
"""@private"""

plate_aliases: dict[str, str] = {
    "暁": "晓",
    "櫻": "樱",
    "菫": "堇",
    "輝": "辉",
    "華": "华",
    "極": "极",
}
"""@private"""


class LevelIndex(Enum):
    BASIC = 0
    ADVANCED = 1
    EXPERT = 2
    MASTER = 3
    ReMASTER = 4


class FCType(Enum):
    APP = 0
    AP = 1
    FCP = 2
    FC = 3


class FSType(Enum):
    SYNC = 0
    FS = 1
    FSP = 2
    FSD = 3
    FSDP = 4


class RateType(Enum):
    SSSP = 0
    SSS = 1
    SSP = 2
    SS = 3
    SP = 4
    S = 5
    AAA = 6
    AA = 7
    A = 8
    BBB = 9
    BB = 10
    B = 11
    C = 12
    D = 13

    @staticmethod
    def _from_achievement(achievement: float) -> "RateType":
        if achievement >= 100.5:
            return RateType.SSSP
        if achievement >= 100:
            return RateType.SSS
        if achievement >= 99.5:
            return RateType.SSP
        if achievement >= 99:
            return RateType.SS
        if achievement >= 98:
            return RateType.SP
        if achievement >= 97:
            return RateType.S
        if achievement >= 94:
            return RateType.AAA
        if achievement >= 90:
            return RateType.AA
        if achievement >= 80:
            return RateType.A
        if achievement >= 75:
            return RateType.BBB
        if achievement >= 70:
            return RateType.BB
        if achievement >= 60:
            return RateType.B
        if achievement >= 50:
            return RateType.C
        return RateType.D


class SongType(Enum):
    STANDARD = "standard"
    DX = "dx"
    UTAGE = "utage"

    @staticmethod
    def _from_id(id: Union[int, str]) -> "SongType":
        id = int(id)
        return SongType.UTAGE if id > 100000 else SongType.DX if id > 10000 else SongType.STANDARD

    def _to_id(self, id: Union[int, str]) -> int:
        id = int(id)
        return id if self == SongType.STANDARD else id + 10000 if self == SongType.DX else id + 100000

    def _to_abbr(self) -> str:
        return "SD" if self == SongType.STANDARD else "DX" if self else "UTAGE"



import hashlib
import json
from pathlib import Path
from typing import TYPE_CHECKING

from maimai_py.models import *

from ..base import IAreaProvider, IItemListProvider

if TYPE_CHECKING:
    from maimai_py.maimai import MaimaiClient


class LocalProvider(IItemListProvider, IAreaProvider):
    """The provider that fetches data from the local storage.

    Most of the data are stored in JSON files in the same directory as this file.
    """

    def _hash(self) -> str:
        current_folder = Path(__file__).resolve().parent
        json_files = sorted(current_folder.glob("*.json"))

        if not json_files:
            return hashlib.md5(b"local").hexdigest()

        combined_content = b""
        for file_path in json_files:
            with open(file_path, "rb") as f:
                combined_content += f.read()

        return hashlib.md5(combined_content).hexdigest()

    def _read_file(self, file_name: str) -> Any:
        current_folder = Path(__file__).resolve().parent
        path = current_folder / f"{file_name}.json"
        if not path.exists():
            raise FileNotFoundError(f"File {path} not found.")
        with open(path, "r", encoding="utf-8") as f:
            return json.load(f)

    def _read_file_dict(self, file_name: str) -> dict:
        obj = self._read_file(file_name)
        if isinstance(obj, dict):
            return obj["data"]
        else:
            raise ValueError(f"File {file_name} is not a dictionary.")

    def _read_file_list(self, file_name: str) -> list:
        obj = self._read_file(file_name)
        if isinstance(obj, list):
            return obj
        else:
            raise ValueError(f"File {file_name} is not a list.")

    async def get_icons(self, client: "MaimaiClient") -> dict[int, PlayerIcon]:
        return {int(k): PlayerIcon(id=int(k), name=v, description=None, genre=None) for k, v in self._read_file_dict("icons").items()}

    async def get_nameplates(self, client: "MaimaiClient") -> dict[int, PlayerNamePlate]:
        return {int(k): PlayerNamePlate(id=int(k), name=v, description=None, genre=None) for k, v in self._read_file_dict("nameplates").items()}

    async def get_frames(self, client: "MaimaiClient") -> dict[int, PlayerFrame]:
        return {int(k): PlayerFrame(id=int(k), name=v, description=None, genre=None) for k, v in self._read_file_dict("frames").items()}

    async def get_partners(self, client: "MaimaiClient") -> dict[int, PlayerPartner]:
        return {int(k): PlayerPartner(id=int(k), name=v) for k, v in self._read_file_dict("partners").items()}

    async def get_charas(self, client: "MaimaiClient") -> dict[int, PlayerChara]:
        return {int(k): PlayerChara(id=int(k), name=v) for k, v in self._read_file_dict("charas").items()}

    async def get_trophies(self, client: "MaimaiClient") -> dict[int, PlayerTrophy]:
        return {int(k): PlayerTrophy(id=int(k), name=v["title"], color=v["rareType"]) for k, v in self._read_file_dict("trophies").items()}

    async def get_areas(self, lang: str, client: "MaimaiClient") -> dict[str, Area]:
        maimai_songs = await client.songs()
        areas = {
            item["id"]: Area(
                id=item["id"],
                name=item["name"],
                comment=item["comment"],
                description=item["description"],
                video_id=item["video_id"],
                characters=[
                    AreaCharacter(
                        name=char["name"],
                        illustrator=char["illustrator"],
                        description1=char["description1"],
                        description2=char["description2"],
                        team=char["team"],
                        props=char["props"],
                    )
                    for char in item["characters"]
                ],
                songs=[
                    AreaSong(
                        id=None,
                        title=song["title"],
                        artist=song["artist"],
                        description=song["description"],
                        illustrator=song["illustrator"],
                        movie=song["movie"],
                    )
                    for song in item["songs"]
                ],
            )
            for item in self._read_file_list(f"areas_{lang}")
        }
        for area in areas.values():
            for song in area.songs:
                maimai_song = await maimai_songs.by_title(song.title)
                song.id = maimai_song.id if maimai_song else None
        return areas
